import numpy as np


def is_pareto_efficient(A: np.ndarray) -> np.ndarray:
    """Find the Pareto-efficient points in a set of objective vectors.

    Args:
        A (2D-array, shape=(samples, dimension)): Objective vectors.

    Returns:
        1D-array of bools: Boolean mask for the Pareto efficient points in A.
    """
    efficient = np.ones(len(A), dtype=bool)
    idx = np.arange(len(A))
    for i, a in enumerate(A):
        if not efficient[i]:
            continue
        # set all *other* efficent points to False, if they are not strictly better in at least one objective
        efficient[efficient] = np.any(A[efficient] < a, axis=1) | (i == idx[efficient])
    return efficient


def pareto_front(A: np.ndarray) -> np.ndarray:
    """Find the Pareto-efficient points in a set of objective vectors.

    Args:
        A (2D-array, shape=(samples, dimension)): Objective vectors.

    Returns:
        2D-array: Pareto efficient points in A.
    """
    return A[is_pareto_efficient(A)]


def crowding_distance(A):
    """Crowding distance indicator.

    The crowding distance is defined for each point in a Pareto front as the average
    side length of the cuboid formed by the neighbouring points.

    Reference:
        [Kalyanmoy Deb (2000) A fast elitist non-dominated sorting genetic algorithm for multi-objective optimization: NSGA-II](https://link.springer.com/chapter/10.1007/3-540-45356-3_83)

    Args:
        A (2D-array): Set of points representing a Pareto front.
            Pareto-efficiency is assumed but not checked.

    Returns:
        array: Crowding distance indicator for each point in the front.
    """
    A = pareto_front(A)
    N, m = A.shape

    # no crowding distance for 2 points
    if N <= 2:
        return np.full(N, np.inf)

    # sort points along each objective
    sort = np.argsort(A, axis=0)
    A = A[sort, np.arange(m)]

    # normalize all objectives
    norm = np.max(A, axis=0) - np.min(A, axis=0)
    A = A / norm
    A[:, norm == 0] = 0  # handle min = max

    # distance to previous and to next point along each objective
    d = np.diff(A, axis=0)
    inf = np.full((1, m), np.inf)
    d0 = np.concatenate([inf, d])
    d1 = np.concatenate([d, inf])

    # TODO: handle cases with duplicate objective values leading to 0 distances

    # cuboid side length = distance between previous and next point
    unsort = np.argsort(sort, axis=0)
    cuboid = d0[unsort, np.arange(m)] + d1[unsort, np.arange(m)]
    return np.mean(cuboid, axis=1)


def generational_distance(
    A: np.ndarray, R: np.ndarray, p: float = 1, clip: bool = True
) -> float:
    r"""Generational distance indicator.

    The generational distance (GD) is defined as the average distance of points in the
    approximate front A to a reference front R.
    .. math:: \mathrm{GD}(A, R) = (\sum\limits_{a \in A} d(a, R)^p)^{1/p} / N_A
    where d(a, R) is the euclidean distance of point a to the reference front, N_A is
    the number of points in A. GD is a measure of convergence.

    Reference:
        [David Van Veldhuizen+ (2000) Multiobjective Evolutionary Algorithms: Analyzing the State-of-the-Art](http://dx.doi.org/10.1162/106365600568158)

    Args:
        A (2D-array): Set of points representing an approximate Pareto front.
        R (2D-array): Set of points representing a reference Pareto front.
        p (int, optional): Order of the p-norm for averaging over distances.
            Defaults to 1, yielding the standard average.
        clip (bool, optional): Flag for using the modfied generational distance, which
            prevents negative values for points that are non-dominated by the reference
            front.

    Returns:
        float: Generational distance indicator.
    """
    A = pareto_front(A)
    distances = A[:, np.newaxis] - R[np.newaxis]
    if clip:
        distances = distances.clip(0, None)
    distances = np.linalg.norm(distances, axis=2).min(axis=1)
    return np.linalg.norm(distances, p) / len(A)


def inverted_generational_distance(A: np.ndarray, R: np.ndarray, p: float = 1) -> float:
    """Inverted generational distance indicator.

    The inverted generational distance (IGD) is defined as the average distance of
    points in the reference front R to an approximate front A.
    IGD is a measure of convergence, spread and distribution of the approximate front.

    Reference:
        [CA. Coello Coello+ (2004) A study of the parallelization of a coevolutionary multi-objective evolutionary algorithm](https://doi.org/10.1007/978-3-540-24694-7_71)

    Args:
        A (2D-array): Set of points representing an approximate Pareto front.
        R (2D-array): Set of points representing a reference Pareto front.
        p (int, optional): Order of the p-norm for averaging over distances.
            Defaults to 1, yielding the standard average.

    Returns:
        float: Inverted generational distance indicator.
    """
    return generational_distance(R, A, p, clip=False)



import numpy as np
import pytest

from opti.constraint import (
    Constraints,
    LinearEquality,
    LinearInequality,
    NChooseK,
    NonlinearEquality,
    NonlinearInequality,
)
from opti.parameter import Categorical, Continuous, Discrete, Parameters
from opti.sampling import (
    constrained_sampling,
    polytope,
    rejection_sampling,
    simplex,
    sobol_sampling,
    sphere,
)


def test_simplex_grid():
    grid = simplex.grid(dimension=3, levels=5)
    assert grid.shape[1] == 3
    assert (grid.sum(axis=1) == 1).all()
    assert (np.unique(grid) == np.array([0, 0.25, 0.5, 0.75, 1])).all()


def test_simplex_sampling():
    points = simplex.sample(dimension=3, n_samples=100)
    assert points.shape == (100, 3)
    assert np.allclose(points.sum(axis=1), 1)
    assert (points >= 0).all()
    assert (points <= 1).all()


def test_sphere_sampling():
    points = sphere.sample(dimension=5, n_samples=1000)
    assert points.shape == (1000, 5)
    assert (np.abs(points) < 1).all()

    points = sphere.sample(dimension=5, n_samples=1000, positive=True)
    assert (points > 0).all()
    assert (points < 1).all()


def test_polytope_chebyshev_center():
    A = np.row_stack([-np.eye(3), np.eye(3)])
    b = np.r_[-np.zeros(3), np.ones(3)]
    x0 = polytope._chebyshev_center(A, b)
    assert np.allclose(x0, [0.5, 0.5, 0.5])


def test_polytope_affine_subspace():
    # x1 + x2 + x3 = 2
    A = np.array([[1, 1, 1]])
    b = np.array([2])
    N, xp = polytope._affine_subspace(A, b)
    Z = np.random.uniform(low=-1, high=1, size=(5, 2))
    X = Z @ N.T + xp
    assert np.allclose(X @ A.T, b)


# def test_polytope_sample():
#     # Sampling from unit-simplex in R^3
#     X = polytope.sample(
#         samples=100,
#         lower=[0, 0, 0],
#         upper=[1, 1, 1],
#         A2=np.array([[1, 1, 1]]),
#         b2=np.array([1]),
#     )
#     assert np.allclose(X.sum(axis=1), 1)
#     assert (X >= 0).all()
#     assert (X <= 1).all()

#     # Sampling from [0, 1]^2 subject to
#     # x1 / 2 + x2 <= 1
#     # 2/3 x1 - x2 <= -0.2
#     A1 = np.array([[1 / 2, 1], [2 / 3, -1]])
#     b1 = np.array([1, -0.2])
#     X = polytope.sample(samples=100, lower=[0, 0], upper=[1, 1], A1=A1, b1=b1)
#     assert np.all(X @ A1.T <= b1)


def test_polytope_sampling():
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(8)])
    constraints = Constraints(
        [
            LinearEquality(names=["x0", "x1", "x2", "x3"], lhs=1, rhs=1),
            LinearEquality(names=["x4", "x5", "x6"], lhs=1, rhs=1),
            LinearInequality(names=["x0", "x1", "x2"], lhs=1, rhs=0.3),
        ]
    )
    X = polytope.polytope_sampling(100, parameters, constraints)
    assert parameters.contains(X).all()
    assert constraints.satisfied(X).all()

    # test without equalities
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints(
        [LinearInequality(names=["x0", "x1", "x2"], lhs=1, rhs=1)]
    )
    X = polytope.polytope_sampling(100, parameters, constraints)
    assert parameters.contains(X).all()
    assert constraints.satisfied(X).all()

    # test without inequalities
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints([LinearEquality(names=["x0", "x1", "x2"], lhs=1, rhs=1)])
    X = polytope.polytope_sampling(100, parameters, constraints)
    assert parameters.contains(X).all()
    assert constraints.satisfied(X).all()


def test_rejection_sampling():
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])

    # test sampling from unit-simplex
    constraints = Constraints(
        [LinearInequality(names=["x0", "x1", "x2"], lhs=1, rhs=1)]
    )
    X = rejection_sampling(100, parameters, constraints)
    assert parameters.contains(X).all()
    assert constraints.satisfied(X).all()

    # test sampling from unit-sphere
    constraints = Constraints([NonlinearInequality("x0**2 + x1**2 + x2**2 - 1")])
    X = rejection_sampling(100, parameters, constraints)
    assert parameters.contains(X).all()
    assert constraints.satisfied(X).all()

    # test max_iters exceeded
    with pytest.raises(Exception):
        rejection_sampling(10000, parameters, constraints, max_iters=1)


def test_constrained_sampling():
    # linear equalities -> polytope sampling
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints([LinearEquality(names=["x0", "x1", "x2"], lhs=1, rhs=1)])
    X = constrained_sampling(10, parameters, constraints)
    assert X.shape == (10, 3)

    # nonlinear inequalities -> rejection sampling
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints([NonlinearInequality("x0**2 + x1**2 + x2**2 - 1")])
    X = constrained_sampling(10, parameters, constraints)
    assert X.shape == (10, 3)

    # non-linear equalities -> error
    parameters = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints([NonlinearEquality("x0**2 + x1**2 + x2**2 - 1")])
    with pytest.raises(Exception):
        constrained_sampling(10, parameters, constraints)

    # linear equalities and non-continuous parameters -> error
    parameters = Parameters(
        [Continuous("x0", domain=[0, 1]), Discrete("x1", [0, 1, 2])]
    )
    constraints = Constraints([NonlinearEquality("x0**2 + x1**2 - 1")])
    with pytest.raises(Exception):
        constrained_sampling(10, parameters, constraints)


def test_constrained_nchoosek():
    # two n-choose-k constraints
    parameters = Parameters([Continuous(f"x{i+1}", domain=[0, 1]) for i in range(10)])
    constraints = Constraints(
        [
            NChooseK(["x1", "x2", "x3", "x4", "x5"], max_active=3),
            NChooseK(["x6", "x7", "x8", "x9", "x10"], max_active=3),
        ]
    )
    X = constrained_sampling(10, parameters, constraints)
    assert constraints.satisfied(X).all()

    # n-choose-k constraint + a particular inequality constraint
    parameters = Parameters([Continuous(f"x{i+1}", domain=[0, 1]) for i in range(3)])
    constraints = Constraints(
        [
            NChooseK(["x1", "x2", "x3"], max_active=2),
            LinearInequality(["x1", "x2", "x3"], lhs=1, rhs=1),
        ]
    )
    X = constrained_sampling(10, parameters, constraints)
    assert constraints.satisfied(X).all()


def test_sobol_sampling():
    parameters = Parameters(
        [
            Continuous("x1", domain=[10, 20]),
            Discrete("x2", domain=[1, 9, 27, 28]),
            Categorical("x3", domain=["A", "B", "C"]),
        ]
    )
    X = sobol_sampling(100, parameters)
    assert np.all(parameters.contains(X))



import json

import numpy as np
import pandas as pd
import pytest
from scipy import stats

import opti
from opti import Categorical, Continuous, Discrete, Problem
from opti.constraint import LinearEquality
from opti.objective import CloseToTarget, Maximize, Minimize
from opti.problems.multi_fidelity import create_multi_fidelity
from opti.problems.noisify import (
    NoiseType,
    _add_noise_to_data,
    noisify_problem,
    noisify_problem_with_gaussian,
    noisify_problem_with_scipy_stats,
)
from opti.sampling import constrained_sampling


def test_single_objective_problems():
    for problem in (
        opti.problems.single.Ackley(),
        opti.problems.single.Himmelblau(),
        opti.problems.single.Rastrigin(),
        opti.problems.single.Rosenbrock(),
        opti.problems.single.Schwefel(),
        opti.problems.single.Sphere(),
        opti.problems.single.Zakharov(),
        opti.problems.single.Zakharov_Categorical(),
        opti.problems.single.Zakharov_Constrained(),
        opti.problems.single.Zakharov_NChooseKConstraint(),
    ):
        # test json-serializability
        json.dumps(problem.to_config())

        # test function evaluations
        x = problem.sample_inputs(10)
        y1 = problem.f(x.values)
        y2 = problem.eval(x)
        assert np.allclose(y1, y2["y"])
        assert np.all(problem.outputs.contains(y2))

        # test the optima
        optima = problem.get_optima()
        px = optima[problem.inputs.names]
        py = optima[problem.outputs.names]
        assert np.allclose(problem.eval(px), py, atol=1e-04)


def test_multi_objective_problems():
    for problem in (
        opti.problems.Qapi1(),
        opti.problems.Hyperellipsoid(),
        opti.problems.Daechert1(),
        opti.problems.Daechert2(),
        opti.problems.Daechert3(),
        opti.problems.Detergent(),
    ):
        json.dumps(problem.to_config())  # test serializing

        # evaluate the function
        x = problem.inputs.sample(10)
        y1 = problem.f(x.values)
        y2 = problem.eval(x)
        assert np.allclose(y1, y2[problem.outputs.names].values)
        assert np.all(problem.outputs.contains(y2))

        # evaluate the constraints
        if problem.constraints:
            problem.constraints.eval(x)
            problem.constraints.satisfied(x)


def test_baking_problems():
    problem = opti.problems.Cake()
    json.dumps(problem.to_config())
    assert set(problem.inputs.names + problem.outputs.names) == set(
        problem.data.columns
    )

    problem = opti.problems.Bread()
    json.dumps(problem.to_config())


def test_dataset_problems():
    for _Problem in (
        opti.problems.Alkox,
        opti.problems.BaumgartnerAniline,
        opti.problems.BaumgartnerBenzamide,
        opti.problems.Benzylation,
        opti.problems.Fullerenes,
        opti.problems.HPLC,
        opti.problems.Photodegradation,
        opti.problems.ReizmanSuzuki,
        opti.problems.Suzuki,
        opti.problems.SnAr,
    ):
        problem = _Problem()
        assert problem.inputs.contains(problem.data[problem.inputs.names]).all()


def test_hyperellipsoid_problem():
    # Hypersphere
    problem = opti.problems.Hyperellipsoid(n=3)
    json.dumps(problem.to_config())
    x = problem.inputs.sample()
    y = problem.eval(x)
    assert problem.outputs.contains(y)

    # sampling from the constrained input space should work
    problem.create_initial_data(10)
    assert len(problem.data) == 10

    # sampling from the Pareto front
    front = problem.get_optima(1)
    assert len(front) == 1
    front = problem.get_optima(10)
    assert len(front) == 10

    # Hyperellipsoid
    with pytest.raises(ValueError):
        opti.problems.Hyperellipsoid(n=3, a=[1, 1])  # a has wrong shape

    problem = opti.problems.Hyperellipsoid(n=2, a=[1000, 1])
    x = problem.inputs.sample()
    y = problem.eval(x)

    front = problem.get_optima(10)
    front.loc[0, "y0"] = -1000
    front.loc[0, "y1"] = -1


def test_detergent():
    p1 = opti.problems.Detergent()
    p1.create_initial_data(1)
    p1.create_initial_data(10)

    p2 = opti.problems.Detergent_OutputConstraint()
    p2.create_initial_data(1)
    p2.create_initial_data(10)

    p2 = opti.problems.Detergent_NChooseKConstraint()
    # p2.create_initial_data(1)  # sampling for n-choose-k constraints not implemented
    # p2.create_initial_data(10)


def test_zdt_problems():
    for problem in (
        opti.problems.ZDT1(n_inputs=20),
        opti.problems.ZDT2(n_inputs=20),
        opti.problems.ZDT3(n_inputs=20),
        opti.problems.ZDT4(n_inputs=20),
        opti.problems.ZDT6(n_inputs=20),
    ):
        json.dumps(problem.to_config())

        x = problem.inputs.sample()
        y = problem.eval(x)
        assert np.all(problem.outputs.contains(y))

        # test obtaining the true pareto front
        opt = problem.get_optima(points=50)
        assert len(opt) == 50


def test_univariate_problems():
    for problem in (
        opti.problems.Line1D(),
        opti.problems.Parabola1D(),
        opti.problems.Sinus1D(),
        opti.problems.Sigmoid1D(),
        opti.problems.Step1D(),
    ):
        x = problem.inputs.sample()
        y = problem.eval(x)
        assert np.all(problem.outputs.contains(y))


def test_mixed_variables_problems():
    problem = opti.problems.DiscreteVLMOP2(5)
    x = problem.inputs.sample(1000)
    y = problem.eval(x)
    assert np.all(problem.outputs.contains(y))

    problem = opti.problems.DiscreteFuelInjector()
    x = problem.inputs.sample(1000)
    y = problem.eval(x)
    assert np.all(problem.outputs.contains(y))


def test_add_noise_to_data():
    cake = opti.problems.Cake()
    cake.f = lambda x: np.zeros((len(x), cake.n_outputs))
    Y = cake.get_Y()

    def noisify(Y):
        return Y

    Y_not_noisy = _add_noise_to_data(Y, [noisify] * Y.shape[1], cake.outputs)
    assert np.allclose(Y, Y_not_noisy)

    def noisify(Y):
        return Y * 0

    # outputs are clipped to their domain bounds after noise is applied
    # this is why the first output ("calories") is set to 300
    Y_not_noisy = _add_noise_to_data(Y, [noisify] * Y.shape[1], cake.outputs)
    assert np.allclose(Y_not_noisy, [300, 0, 0])


def test_noisify_problem():
    cake = opti.problems.Cake()
    cake.outputs["calories"].domain = [0, 600]
    cake.outputs["calories"].low = 0

    def test(rvs, rv_kwargs):
        for noise_type, neutral_element in (
            (NoiseType.ADDITIVE, 0),
            (NoiseType.MULTIPLICATIVE, 1),
            (None, 0),
        ):
            cake.f = lambda x: np.full((len(x), cake.n_outputs), neutral_element)

            # create a large enough data set for a KS test
            X = cake.inputs.sample(1000)
            Y = pd.DataFrame(
                columns=cake.outputs.names,
                data=np.full((1000, cake.n_outputs), neutral_element),
            )
            cake.data = pd.concat([X, Y], axis=1)

            if noise_type is None:
                noisy_cake = noisify_problem_with_scipy_stats(
                    cake, rvs, rv_kwargs=rv_kwargs
                )
            else:
                noisy_cake = noisify_problem_with_scipy_stats(
                    cake, rvs, rv_kwargs=rv_kwargs, noise_types=noise_type
                )

            # test that both data and function are noisified
            noisy_Y = np.row_stack([noisy_cake.get_Y(), noisy_cake.f(X.values)])

            rv_kwargs_ref = rv_kwargs if isinstance(rv_kwargs, list) else [rv_kwargs]
            rvs_ref = rvs if isinstance(rvs, list) else [rvs]
            for col, rv, rvk in zip(noisy_Y.T, rvs_ref, rv_kwargs_ref):
                reference = rv.rvs(**rvk, size=col.size)
                # since we clip noisy problems to the bounds we also clip the reference
                reference[reference < 0] = 0
                s, _ = stats.ks_2samp(col, reference)
                assert s < 0.1

    test(stats.truncnorm, {"a": 0.1, "b": 2})
    test(
        [stats.truncnorm, stats.norm, stats.norm],
        [{"a": 0.1, "b": 2}, {"loc": 0.1, "scale": 0.3}, {"loc": -0.1, "scale": 0.1}],
    )


def test_noisify_categorical_discrete_outputs():
    problem = Problem(
        inputs=[Continuous("i1", [0, 1])],
        outputs=[
            Categorical("o1", [1, 2, 7]),
            Discrete("o2", [0, 1]),
            Discrete("o3", [0, 1, 2]),
        ],
        f=lambda x: np.array([[1, 1, 2]]),
    )

    def noisify(Y):
        return Y + 1

    noisy_cat_disc_problem = noisify_problem(problem, noisify)

    # Cat. dummy value will be increased by 1, discrete will be out of bounds after + 1
    assert np.allclose(noisy_cat_disc_problem.f(1), np.array([[2, 1, 2]]))


def test_noisify_problem_with_gaussian():
    n_samples = 5000
    mu = 0.1
    sigma = 0.05
    reference = np.clip(stats.norm.rvs(loc=mu, scale=sigma, size=n_samples), 0, 1)

    zdt2 = opti.problems.ZDT2()
    zdt2.create_initial_data(n_samples)
    zdt2_gaussian = noisify_problem_with_gaussian(zdt2, mu=mu, sigma=sigma)
    Y_gaussian = zdt2_gaussian.get_Y()
    Y = zdt2.get_Y()
    Y_noise = Y_gaussian - Y
    for col in Y_noise.T:
        s, _ = stats.ks_2samp(col, reference)
        assert s < 0.1


def test_multi_fidelity():
    problem = Problem(
        inputs=[Continuous("i", [0, 1])],
        outputs=[
            Continuous("o", [0, 1]),
        ],
        f=lambda x: np.array([[0.5]]),
    )

    def second_fidelity(_):
        return np.array([[1.0]])

    def first_predicate(counter, _):
        return counter % 3 == 0

    def second_predicate(counter, _):
        return not first_predicate(counter, _)

    mf_problem = create_multi_fidelity(
        problem, [problem.f, second_fidelity], [first_predicate, second_predicate]
    )
    for i in range(100):
        y = mf_problem.f(None)
        if i % 3 == 0:
            assert y[0, 0] == 0.5
        else:
            assert y[0, 0] == 1.0

    n_inputs = 5
    zdt1 = opti.problems.ZDT1(n_inputs=n_inputs)
    first_fidelity = noisify_problem_with_gaussian(zdt1, mu=0, sigma=0.01)
    second_fidelity = noisify_problem_with_gaussian(zdt1, mu=0.1, sigma=0.5)

    def pred_second(counter, _):
        return counter % 3 == 0

    def pred_first(counter, _):
        return not pred_second(counter, _)

    mf_zdt = create_multi_fidelity(
        zdt1, [first_fidelity.f, second_fidelity.f], [pred_first, pred_second]
    )

    n_samples = 3000

    Y_mf = -np.ones((n_samples, 2))
    Y = -np.ones((n_samples, 2))
    X = constrained_sampling(n_samples, mf_zdt.inputs, mf_zdt.constraints).values
    for i in range(n_samples):
        # needs to be in a loop so that the counter of mf_zdt is increased
        Y_mf[i, :] = mf_zdt.f(X[i].reshape(1, n_inputs))
        Y[i, :] = zdt1.f(X[i].reshape(1, n_inputs))

    f2_inds = np.array([pred_second(i, None) for i in range(n_samples)])
    f1_inds = (1 - f2_inds) > 0

    Y_1_noisy = Y_mf[f1_inds]
    Y_2_noisy = Y_mf[f2_inds]
    Y_1 = Y[f1_inds]
    Y_2 = Y[f2_inds]
    Y_1_noise = np.clip(Y_1_noisy - Y_1, 0, 1)
    Y_2_noise = np.clip(Y_2_noisy - Y_2, 0, 1)
    reference_1 = np.clip(stats.norm.rvs(loc=0, scale=0.01, size=np.sum(f2_inds)), 0, 1)
    reference_2 = np.clip(
        stats.norm.rvs(loc=0.1, scale=0.5, size=np.sum(f1_inds)), 0, 1
    )

    for Y_noise, ref in zip([Y_1_noise, Y_2_noise], [reference_1, reference_2]):
        for col in Y_noise.T:
            s, _ = stats.ks_2samp(col, ref)
            assert s < 0.1


def test_sanitize_problem():
    def _test(original):
        sanitized = opti.problems.sanitize_problem(original)

        # check inputs
        for p1, p2 in zip(original.inputs, sanitized.inputs):
            assert type(p1) == type(p2)
            if isinstance(p2, Continuous):
                p2.bounds == (0, 1)  # continuous input bounds are [0, 1]
            assert p2.name.startswith("input")

        # check outputs
        for p1, p2 in zip(original.outputs, sanitized.outputs):
            assert type(p1) == type(p2)
            if isinstance(
                p2, Continuous
            ):  # continuous output bounds are [0, 1] if specified and [-inf, inf] if not
                if np.isneginf(p1.low):
                    assert np.isneginf(p2.low)
                else:
                    assert np.isclose(p2.low, 0)
                if np.isinf(p1.high):
                    assert np.isinf(p2.high)
                else:
                    assert np.isclose(p2.high, 1)
            assert p2.name.startswith("output")

        # check objectives
        for (p1, p2) in zip(original.objectives, sanitized.objectives):
            assert type(p1) == type(p2)
            assert p2.name.startswith("output")
            assert np.isfinite(p2.target)

        # check constraints
        if original.constraints is not None:
            for c1, c2 in zip(original.constraints, sanitized.constraints):
                # sanitizing should not modify the original problem
                assert c1.names != c2.names

        # check data
        assert original.data.shape == sanitized.data.shape
        assert set(sanitized.data.columns) == set(
            sanitized.inputs.names + sanitized.outputs.names
        )
        assert sanitized.inputs.contains(sanitized.data[sanitized.inputs.names]).all()
        assert sanitized.constraints.satisfied(sanitized.data).all()

        # problem.f is dropped
        assert not hasattr(sanitized, "f")

    _test(opti.problems.Cake())

    _test(opti.problems.Photodegradation())

    z = opti.problems.Zakharov_Constrained()
    z.create_initial_data(n_samples=1000)
    _test(z)

    tp = Problem(
        inputs=[
            Continuous("secret_ingredient_1", domain=[0, 150]),
            Continuous("secret_ingredient_2", domain=[0, 50]),
            Continuous("secret_ingredient_3", domain=[0, 34.3]),
        ],
        outputs=[
            Continuous("ultimate_goal", domain=[0, np.linalg.norm([150, 50, 34.3])]),
            Continuous("mega_goal", domain=[0, np.sum(np.square([150, 50, 34.3]))]),
            Continuous(
                "super_goal", domain=[0, np.linalg.norm([150, 50, 34.3]) ** 0.5]
            ),
        ],
        objectives=[
            CloseToTarget("ultimate_goal", target=100),
            Minimize("mega_goal", target=1000),
            Maximize("super_goal", target=10),
        ],
        constraints=[
            LinearEquality(
                ["secret_ingredient_1", "secret_ingredient_2", "secret_ingredient_3"],
                lhs=np.array([3.5, 4.5, 2]),
                rhs=153.1,
            )
        ],
        f=lambda x: np.stack(
            [
                np.linalg.norm(x, axis=1),
                np.sum(np.square(x), axis=1),
                np.linalg.norm(x, axis=1) ** 0.5,
            ],
            axis=1,
        ),
    )
    tp.create_initial_data(n_samples=100)
    _test(tp)

    empty = Problem([], [])  # this should not be possible
    with pytest.raises(TypeError):
        opti.problems.sanitize_problem(empty)



import os

import pandas as pd

from opti.constraint import LinearEquality, NChooseK
from opti.objective import Maximize, Minimize
from opti.parameter import Categorical, Continuous
from opti.problem import Problem

cdir = os.path.dirname(os.path.realpath(__file__))


def get_data(fname: str):
    return pd.read_csv(f"{cdir}/data/{fname}")


class Alkox(Problem):
    def __init__(self):
        super().__init__(
            name="Alkox",
            inputs=[
                Continuous("residence_time", domain=[0.05, 1]),
                Continuous("ratio", domain=[0.5, 10]),
                Continuous("concentration", domain=[2, 8]),
                Continuous("temperature", domain=[6, 8]),
            ],
            outputs=[Continuous("conversion")],
            objectives=[Maximize("conversion")],
            data=get_data("alkox.csv"),
        )


class BaumgartnerAniline(Problem):
    def __init__(self):
        super().__init__(
            name="Baumgartner 2019 - Aniline Cross-Coupling",
            inputs=[
                Categorical("catalyst", domain=["tBuXPhos", "tBuBrettPhos", "AlPhos"]),
                Categorical("base", domain=["TEA", "TMG", "BTMG", "DBU"]),
                Continuous("base_equivalents", domain=[1.0, 2.5]),
                Continuous("temperature", domain=[30, 100]),
                Continuous("residence_time", domain=[60, 1800]),
            ],
            outputs=[Continuous("yield", domain=[0, 1])],
            objectives=[Maximize("yield")],
            data=get_data("baumgartner_aniline.csv"),
        )


class BaumgartnerBenzamide(Problem):
    def __init__(self):
        super().__init__(
            name="Baumgartner 2019 - Benzamide Cross-Coupling",
            inputs=[
                Categorical("catalyst", domain=["tBuXPhos", "tBuBrettPhos"]),
                Categorical("base", domain=["TMG", "BTMG", "DBU", "MTBD"]),
                Continuous("base_equivalents", domain=[1.0, 2.1]),
                Continuous("temperature", domain=[30, 100]),
                Continuous("residence_time", domain=[60, 1850]),
            ],
            outputs=[Continuous("yield", domain=[0, 1])],
            objectives=[Maximize("yield")],
            data=get_data("baumgartner_benzamide.csv"),
        )


class Benzylation(Problem):
    def __init__(self):
        super().__init__(
            name="Benzylation",
            inputs=[
                Continuous("flow_rate", domain=[0.2, 0.4]),
                Continuous("ratio", domain=[1.0, 5.0]),
                Continuous("solvent", domain=[0.5, 1.0]),
                Continuous("temperature", domain=[110.0, 150.0]),
            ],
            outputs=[Continuous("impurity")],
            objectives=[Minimize("impurity")],
            data=get_data("benzylation.csv"),
        )


class Fullerenes(Problem):
    def __init__(self):
        super().__init__(
            name="Fullerenes",
            inputs=[
                Continuous("reaction_time", domain=[3.0, 31.0]),
                Continuous("sultine", domain=[1.5, 6.0]),
                Continuous("temperature", domain=[100.0, 150.0]),
            ],
            outputs=[Continuous("product")],
            objectives=[Maximize("product")],
            data=get_data("fullerenes.csv"),
        )


class HPLC(Problem):
    def __init__(self):
        super().__init__(
            name="HPLC",
            inputs=[
                Continuous("sample_loop", domain=[0.0, 0.08]),
                Continuous("additional_volume", domain=[0.0, 0.06]),
                Continuous("tubing_volume", domain=[0.1, 0.9]),
                Continuous("sample_flow", domain=[0.5, 2.5]),
                Continuous("push_speed", domain=[80.0, 150]),
                Continuous("wait_time", domain=[0.5, 10.0]),
            ],
            outputs=[Continuous("peak_area")],
            objectives=[Maximize("peak_area")],
            data=get_data("hplc.csv"),
        )


class Photodegradation(Problem):
    def __init__(self):
        super().__init__(
            name="Photodegradation",
            inputs=[
                Continuous("PCE10", domain=[0, 1]),
                Continuous("WF3", domain=[0, 1]),
                Continuous("P3HT", domain=[0, 1]),
                Continuous("PCBM", domain=[0, 1]),
                Continuous("oIDTBR", domain=[0, 1]),
            ],
            outputs=[Continuous("degradation")],
            objectives=[Minimize("degradation")],
            constraints=[
                LinearEquality(
                    ["PCE10", "WF3", "P3HT", "PCBM", "oIDTBR"], rhs=1, lhs=1
                ),
                NChooseK(["PCE10", "WF3"], max_active=1),
            ],
            data=get_data("photodegradation.csv"),
        )


class ReizmanSuzuki(Problem):
    def __init__(self, case=1):
        assert case in [1, 2, 3, 4]
        super().__init__(
            name=f"Reizman 2016 - Suzuki Case {case}",
            inputs=[
                Categorical(
                    "catalyst",
                    domain=[
                        "P1-L1",
                        "P2-L1",
                        "P1-L2",
                        "P1-L3",
                        "P1-L4",
                        "P1-L5",
                        "P1-L6",
                        "P1-L7",
                    ],
                ),
                Continuous("t_res", domain=[60, 600]),
                Continuous("temperature", domain=[30, 110]),
                Continuous("catalyst_loading", domain=[0.496, 2.515]),
            ],
            outputs=[
                Continuous("ton", domain=[0, 100]),
                Continuous("yield", domain=[0, 100]),
            ],
            objectives=[Maximize("ton"), Maximize("yield")],
            data=get_data(f"reizman_suzuki{case}.csv"),
        )


class SnAr(Problem):
    def __init__(self):
        super().__init__(
            name="SnAr",
            inputs=[
                Continuous("residence_time", domain=[0.5, 2.0]),
                Continuous("ratio", domain=[1.0, 5.0]),
                Continuous("concentration", domain=[0.1, 0.5]),
                Continuous("temperature", domain=[60.0, 140.0]),
            ],
            outputs=[Continuous("impurity")],
            objectives=[Minimize("impurity")],
            data=get_data("snar.csv"),
        )


class Suzuki(Problem):
    def __init__(self):
        super().__init__(
            name="Suzuki",
            inputs=[
                Continuous("temperature", domain=[75.0, 90.0]),
                Continuous("pd_mol", domain=[0.5, 5.0]),
                Continuous("arbpin", domain=[1.0, 1.8]),
                Continuous("k3po4", domain=[1.5, 3.0]),
            ],
            outputs=[Continuous("yield")],
            objectives=[Maximize("yield")],
            data=get_data("suzuki.csv"),
        )



import warnings
from copy import deepcopy
from enum import Enum
from typing import Tuple

import numpy as np
import pandas as pd

from opti.constraint import LinearEquality, LinearInequality, NChooseK
from opti.parameter import Continuous, Parameters
from opti.problem import Problem


def _normalize_parameters_data(
    data: pd.DataFrame, params: Parameters
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """Transform data from the domain bounds to the unit range"""

    lo = params.bounds.loc["min"]
    hi = params.bounds.loc["max"]

    lo[lo == -np.inf] = data[lo[lo == -np.inf].index].min(axis=0)
    hi[hi == np.inf] = data[hi[hi == np.inf].index].max(axis=0)

    data = (data[params.names] - lo) / (hi - lo)
    min_vals = lo.values
    delta_vals = (hi - lo).values

    return data, min_vals, delta_vals


class InOrOut(str, Enum):
    IN = "input"
    OUT = "output"


def _sanitize_params(parameters: Parameters, in_or_out: InOrOut):
    """Sets parameter boundaries to 0 and 1 and replaces parameter names"""
    sanitized = []
    for i, p in enumerate(parameters):
        name = f"{in_or_out}_{i}"
        if isinstance(p, Continuous):
            low = -np.inf if np.isneginf(p.low) else 0
            high = np.inf if np.isinf(p.high) else 1
            domain = [low, high]
        else:
            raise TypeError("currently only continuous parameters are supported")
        sanitized.append(type(p)(name, domain))
    return Parameters(sanitized)


def sanitize_problem(problem: Problem):
    """
    This creates a transformation of the problem with sanitized data. Thereby, we try
    to preserve relationships between inputs, outputs, and objectives.

    More precisely, the resulting problem has the following properties:
    - Inputs are named `input_0`, `input_1`, .... Outputs are named analogously.
    - The data is scaled per feature to `[0, 1]`.
    - Coefficients of linear constraints are adapted to the data scaling.
    - Models and evaluatons in terms of `f` are dropped if there are any.

    Currently unsuported are problems with
    - discrete or categorical variables,
    - input constraints that are neither linear nor n-choose-k, or
    - output constraints.

    Args:
        problem: to be sanitized

    Raises:
        TypeError: in case there are unsupported constraints, data is None, or there are output constraints

    Returns:
        Problem instance with sanitized labels and normalized data
    """
    if problem.data is None:
        raise TypeError("we cannot sanitize a problem without data")
    if problem.output_constraints is not None:
        raise TypeError("output constraints are currently not supported")
    if getattr(problem, "f", None) is not None:
        warnings.warn("f is not sanitized but dropped")
    if problem.models is not None:
        warnings.warn("models are not sanitized but dropped")

    inputs = _sanitize_params(problem.inputs, InOrOut.IN)
    input_name_map = {pi.name: i.name for pi, i in zip(problem.inputs, inputs)}
    normalized_in_data, xmin, Δx = _normalize_parameters_data(
        problem.data, problem.inputs
    )
    outputs = _sanitize_params(problem.outputs, InOrOut.OUT)
    output_name_map = {pi.name: i.name for pi, i in zip(problem.outputs, outputs)}
    normalized_out_data, ymin, Δy = _normalize_parameters_data(
        problem.data, problem.outputs
    )

    normalized_in_data.columns = inputs.names
    normalized_out_data.columns = outputs.names
    normalized_data = pd.concat([normalized_in_data, normalized_out_data], axis=1)
    normalized_data.reset_index(inplace=True, drop=True)

    objectives = deepcopy(problem.objectives)
    for obj in objectives:
        sanitized_name = output_name_map[obj.name]
        i = outputs.names.index(sanitized_name)
        obj.name = sanitized_name
        obj.parameter = sanitized_name
        obj.target = (obj.target - ymin[i]) / Δy[i]
        if hasattr(obj, "tolerance"):
            obj.tolerance /= Δy[i]

    constraints = deepcopy(problem.constraints)
    if constraints is not None:
        for c in constraints:
            c.names = [input_name_map[n] for n in c.names]
            if isinstance(c, (LinearEquality, LinearInequality)):
                c.lhs = (c.lhs + xmin) * Δx
                if c.rhs > 1e-5:
                    c.lhs = c.lhs / c.rhs
                    c.rhs = 1.0
            elif isinstance(c, NChooseK):
                pass
            else:
                raise TypeError(
                    "sanitizer only supports linear and n-choose-k constraints"
                )

    normalized_problem = Problem(
        inputs=inputs,
        outputs=outputs,
        objectives=objectives,
        constraints=constraints,
        data=normalized_data,
    )
    return normalized_problem



import numbers
import pprint
from typing import Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

_CAT_SEP = "§"


class Parameter:
    """Parameter base class."""

    def __init__(self, name: str, domain: Sequence, type: str = None, **kwargs):
        self.name = name
        self.domain = domain
        self.type = type
        self.extra_fields = kwargs

    def to_config(self) -> Dict:
        """Return a json-serializable configuration dict."""
        conf = dict(name=self.name, type=self.type, domain=self.domain)
        conf.update(self.extra_fields)
        return conf


class Continuous(Parameter):
    """Parameter that can take on any real value in the specified domain.

    Attributes:
        name (str): name of the parameter
        domain (list): [lower bound, upper bound]
    """

    def __init__(
        self,
        name: str,
        domain: Optional[Sequence] = None,
        **kwargs,
    ):
        if domain is None:
            domain = [-np.inf, np.inf]
        else:
            if len(domain) != 2:
                raise ValueError("domain needs to consist of two values [low, high]")
        # convert None to inf (json doesn't support infinity)
        low = -np.inf if domain[0] is None else domain[0]
        high = np.inf if domain[1] is None else domain[1]
        if high < low:
            raise ValueError(f"lower bound {low} must be less than upper bound {high}")
        self.low = low
        self.high = high
        super().__init__(name=name, domain=[low, high], type="continuous", **kwargs)

    def __repr__(self):
        if np.isfinite(self.low) or np.isfinite(self.high):
            return f"Continuous(name='{self.name}', domain={self.domain})"
        else:
            return f"Continuous(name='{self.name}')"

    @property
    def bounds(self) -> Tuple[float, float]:
        """Return the domain bounds."""
        return self.low, self.high

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        return (self.low <= point) & (point <= self.high)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with values clipped to parameter bounds.
        """
        return np.clip(point, self.low, self.high)

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        low = max(self.low, np.finfo(np.float32).min)
        high = min(self.high, np.finfo(np.float32).max)
        return pd.Series(name=self.name, data=np.random.uniform(low, high, n))

    def to_config(self) -> Dict:
        """Return a json-serializable configuration dict."""
        low = None if np.isinf(self.low) else float(self.low)
        high = None if np.isinf(self.high) else float(self.high)
        conf = dict(name=self.name, type=self.type, domain=[low, high])
        conf.update(self.extra_fields)
        return conf

    def to_unit_range(self, points):
        """Transform points to the unit range: [low, high] -> [0, 1]."""
        if np.isclose(self.low, self.high):
            return points
        else:
            return (points - self.low) / (self.high - self.low)

    def from_unit_range(self, points):
        """Transform points from the unit range: [0-1] -> [low, high]."""
        if np.isclose(self.low, self.high):
            return np.ones_like(points) * self.high
        else:
            return points * (self.high - self.low) + self.low


class Discrete(Parameter):
    """Discrete parameter (ordinal scale).

    Attributes:
        name (str): name of the parameter
        domain (list): list of possible numeric values
    """

    def __init__(self, name: str, domain: Sequence, **kwargs):
        if len(domain) < 1:
            raise ValueError("domain must contain at least 1 value")
        for d in domain:
            if not isinstance(d, numbers.Number):
                raise ValueError("domain contains non-numeric values")
        if len(set(domain)) != len(domain):
            raise ValueError("domain contains duplicates")
        # convert to python-native dtype to ensure json-serializability
        domain = np.array(np.sort(domain)).tolist()
        self.low = min(domain)
        self.high = max(domain)
        super().__init__(name, domain, type="discrete", **kwargs)

    def __repr__(self):
        return f"Discrete(name='{self.name}', domain={self.domain})"

    @property
    def bounds(self) -> Tuple[float, float]:
        """Return the domain bounds."""
        return self.low, self.high

    def is_integer(self):
        """Check if the domain is an integer range, such as [1, 2, 3]"""
        if isinstance(self.low, int) and isinstance(self.high, int):
            return self.domain == list(range(self.low, self.high + 1))
        else:
            return False

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        if not np.isscalar(point):
            point = np.array(point)
        return np.isin(point, self.domain)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with values clipped to parameter bounds.
        """
        if np.isscalar(point):
            i = np.argmin(np.abs(np.array(self.domain) - point))
            return self.domain[i]
        closest = [np.argmin(np.abs(np.array(self.domain) - p)) for p in point]
        rounded = np.array(self.domain)[closest]
        if isinstance(point, np.ndarray):
            return rounded
        elif isinstance(point, pd.Series):
            return pd.Series(name=self.name, data=rounded, index=point.index)
        elif isinstance(point, pd.DataFrame):
            return pd.DataFrame({self.name: rounded}, index=point.index)

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        return pd.Series(name=self.name, data=np.random.choice(self.domain, n))

    def to_unit_range(self, points):
        """Transform points to the unit range.

        Note, if the given points are not inside the domain of the parameter, the
        transformed will not be inside the unit range.
        """
        if np.isclose(self.low, self.high):
            return points
        else:
            return (points - self.low) / (self.high - self.low)


class Categorical(Parameter):
    """Categorical parameter (nominal scale, values cannot be put into order).

    Attributes:
        name (str): name of the parameter
        domain (list): list possible values
    """

    def __init__(self, name: str, domain: Sequence, **kwargs):
        if not isinstance(domain, list):
            raise ValueError("domain must be of type list")
        if len(domain) < 2:
            raise ValueError("domain must a least contain 2 values")
        if len(set(domain)) != len(domain):
            raise ValueError("domain contains duplicates")
        super().__init__(name, domain, type="categorical", **kwargs)

    def __repr__(self):
        return f"Categorical(name='{self.name}', domain={self.domain})"

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        if not np.isscalar(point):
            point = np.array(point)
        return np.isin(point, self.domain)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point`.
        """
        if not np.all(self.contains(point)):
            raise ValueError("cannot round categorical variable")
        return point

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        return pd.Series(name=self.name, data=np.random.choice(self.domain, n))

    def to_onehot_encoding(self, points: pd.Series) -> pd.DataFrame:
        """Convert points to a one-hot encoding."""
        return pd.DataFrame(
            {f"{self.name}{_CAT_SEP}{c}": points == c for c in self.domain}, dtype=float
        )

    def from_onehot_encoding(self, points: pd.DataFrame) -> pd.Series:
        """Convert points back from one-hot encoding."""
        cat_cols = [f"{self.name}{_CAT_SEP}{c}" for c in self.domain]
        if np.any([c not in cat_cols for c in points.columns]):
            raise ValueError(
                f"Column names don't match categorical levels: {points.columns}, {cat_cols}"
            )
        s = points.idxmax(1).str.split(_CAT_SEP, expand=True)[1]
        s.name = self.name
        return s

    def to_dummy_encoding(self, points: pd.Series) -> pd.DataFrame:
        """Convert points to a dummy-hot encoding, dropping the first categorical level."""
        return pd.DataFrame(
            {f"{self.name}{_CAT_SEP}{c}": points == c for c in self.domain[1:]},
            dtype=float,
        )

    def from_dummy_encoding(self, points: pd.DataFrame) -> pd.Series:
        """Convert points back from dummy encoding."""
        cat_cols = [f"{self.name}{_CAT_SEP}{c}" for c in self.domain]
        if np.any([c not in cat_cols[1:] for c in points.columns]):
            raise ValueError(
                f"Column names don't match categorical levels: {points.columns}, {cat_cols}"
            )
        points = points.copy()
        points[cat_cols[0]] = 1 - points[cat_cols[1:]].sum(axis=1)
        s = points.idxmax(1).str.split(_CAT_SEP, expand=True)[1]
        s.name = self.name
        return s

    def to_label_encoding(self, points: pd.Series) -> pd.Series:
        """Convert points to label-encoding."""
        enc = pd.Series(range(len(self.domain)), index=list(self.domain))
        s = enc[points]
        s.index = points.index
        s.name = self.name
        return s

    def from_label_encoding(self, points: pd.Series) -> pd.Series:
        """Convert points back from label-encoding."""
        enc = np.array(self.domain)
        return pd.Series(enc[points], index=points.index)


def make_parameter(
    name: str,
    type: str,
    domain: Optional[Sequence] = None,
    **kwargs,
):
    """Make a parameter object from a configuration

    p = make_parameter(**config)

    Args:
        type (str): "continuous", "discrete" or "categorical"
        name (str): Name of the parameter
        domain (list): Domain, e.g, [0, 1], [1, 2.5, 5] or ["A", "B", "C"]
    """
    parameter = {
        "continuous": Continuous,
        "discrete": Discrete,
        "categorical": Categorical,
    }[type.lower()]
    if domain is None and parameter is not Continuous:
        raise ValueError(f"domain not given for parameter {name}")

    return parameter(name=name, domain=domain, **kwargs)


class Parameters:
    """Set of parameters representing either the input or the output space."""

    def __init__(self, parameters):
        """
        It can be constructed either from a list / tuple (of at least one) Parameter objects
        ```
        Space([
            Continuous(name="foo", domain=[1, 10]),
            Discrete(name="bar", domain=[1, 2, 3, 4]),
            Categorical(name="baz", domain=["A", "B", 3]),
        ])
        ```
        or from a list / tuple of dicts
        ```
        Space([
            {"name": "foo", "type": "continuous", "domain": [1, 10]},
            {"name": "bar", "type": "discrete", "domain": [1, 2, 3, 4]},
            {"name": "baz", "type": "categorical", "domain": ["A", "B", 3]},
            {"name": "baz", "type": "categorical", "domain": ["A", "B", 3], extra="info"},
        ])
        ```
        In particular, Parameters().__init__ and the to_config method of each Parameter type are inverses.
        Parameters(conf).to_config() == conf
        """
        if not isinstance(parameters, (list, tuple)):
            raise TypeError("Space expects a list or tuple of parameters.")

        self.parameters = {}
        for d in parameters:
            if not isinstance(d, Parameter):
                d = make_parameter(**d)
            if d.name in self.parameters:
                raise ValueError(f"Duplicate parameter name {d.name}")
            self.parameters[d.name] = d

    def __repr__(self):
        return "Parameters(\n" + pprint.pformat(list(self.parameters.values())) + "\n)"

    def __iter__(self):
        return iter(self.parameters.values())

    def __getitem__(self, name):
        return self.parameters[name]

    def __len__(self):
        return len(self.parameters)

    def __add__(self, other):
        parameters = list(self.parameters.values()) + list(other.parameters.values())
        return Parameters(parameters)

    @property
    def names(self):
        return list(self.parameters.keys())

    @property
    def bounds(self) -> pd.DataFrame:
        """Return the parameter bounds."""
        for param in self:
            if isinstance(param, Categorical):
                raise TypeError(
                    "Contains categorical parameters which are not bounded."
                )
        return pd.DataFrame({p.name: p.bounds for p in self}, index=["min", "max"])

    def contains(self, points: pd.DataFrame) -> pd.Series:
        """Check if points are inside the space in each parameter."""
        b = np.stack([self[k].contains(v) for k, v in points.iteritems()], axis=-1)
        return b.all(axis=-1)

    def round(self, points: pd.DataFrame) -> pd.DataFrame:
        """Round points to the closest contained values."""
        return pd.concat([self[k].round(v) for k, v in points.iteritems()], axis=1)

    def sample(self, n: int = 1) -> pd.DataFrame:
        """Draw uniformly distributed random samples."""
        return pd.concat([param.sample(n) for param in self], axis=1)

    def transform(
        self,
        points: pd.DataFrame,
        continuous: str = "none",
        discrete: str = "none",
        categorical: str = "onehot-encode",
    ) -> pd.DataFrame:
        transformed = []
        for p in self:
            s = points[p.name]
            if isinstance(p, Continuous):
                if continuous == "none":
                    transformed.append(s)
                elif continuous == "normalize":
                    transformed.append(p.to_unit_range(s))
                else:
                    raise ValueError(f"Unknown continuous transform {continuous}")
            if isinstance(p, Discrete):
                if discrete == "none":
                    transformed.append(s)
                elif discrete == "normalize":
                    transformed.append(p.to_unit_range(s))
                else:
                    raise ValueError(f"Unknown discrete transform {continuous}")
            if isinstance(p, Categorical):
                if categorical == "none":
                    transformed.append(s)
                elif categorical == "onehot-encode":
                    transformed.append(p.to_onehot_encoding(s))
                elif categorical == "dummy-encode":
                    transformed.append(p.to_dummy_encoding(s))
                elif categorical == "label-encode":
                    transformed.append(p.to_label_encoding(s))
                else:
                    raise ValueError(f"Unknown categorical transform {continuous}")
        return pd.concat(transformed, axis=1)

    def to_config(self) -> List[Dict]:
        """Configuration of the parameter space."""
        return [param.to_config() for param in self.parameters.values()]



import json
from unittest import TestCase

import numpy as np
import pandas as pd
import pytest

from opti.parameter import Categorical, Continuous, Discrete, Parameters, make_parameter


def test_arbitrary_parameters():
    # Issue #60
    # saves as {}} when no extra fields
    p = make_parameter(name="foo", type="continuous", domain=[0, 1])
    assert len(p.extra_fields) == 0

    # saves extra fields as a dictionary, of expected form. Independent of type.
    conf_conf = []
    for key, value in {
        "continuous": [0, 1],
        "discrete": [1, 2, 5],
        "categorical": ["A", "B", 3],
    }.items():
        e_dict = {"extra": True, "info": [1, 2, 3], "here": {"a": 1, "b": 2}}
        conf = {"name": "foo" + key, "type": key, "domain": value}
        conf.update(e_dict)
        conf_conf.append(conf)
        p = make_parameter(
            name="foo" + key,
            type=key,
            domain=value,
            extra=True,
            info=[1, 2, 3],
            here={"a": 1, "b": 2},
        )
        TestCase().assertDictEqual(p.extra_fields, e_dict)
        TestCase().assertDictEqual(p.to_config(), conf)

    # test that the parameter constructors and to_config are inverses
    [
        TestCase().assertDictEqual(Parameters(conf_conf).to_config()[i], conf_conf[i])
        for i in range(len(conf_conf))
    ]
    # saves extra fields as a dictionary, of expected form


def test_make_parameter():
    p = make_parameter(name="foo", type="continuous", domain=[0, 1])
    assert isinstance(p, Continuous)
    p = make_parameter(name="foo", type="discrete", domain=[1, 2, 5])
    assert isinstance(p, Discrete)
    p = make_parameter(name="foo", type="categorical", domain=["A", "B", 3])
    assert isinstance(p, Categorical)

    # no bounds
    p = make_parameter(name="foo", type="continuous")
    assert isinstance(p, Continuous)
    with pytest.raises(ValueError):
        make_parameter(name="foo", type="discrete")
    with pytest.raises(ValueError):
        make_parameter(name="foo", type="categorical")


class TestContinuous:
    def test_init(self):
        for domain in ((1, 10), [1, 10], np.array([1, 10])):
            p = Continuous("x", domain=domain)
            assert p.name == "x"
            assert p.bounds == (1, 10)
            assert p.domain == [1, 10]

    def test_checks(self):
        with pytest.raises(ValueError):
            Continuous("x", domain=[20, 1])
        with pytest.raises(ValueError):
            Continuous("x", domain=[1, 2, 3])

    def test_config(self):
        p = Continuous("x", domain=[1, 10])
        conf = p.to_config()
        assert conf["type"] == "continuous"
        assert conf["name"] == "x"
        assert conf["domain"] == [1, 10]
        json.dumps(conf)

        # json-compliance: inf -> None
        p = Continuous("x")
        conf = p.to_config()
        assert conf["domain"] == [None, None]

    def test_config_int32(self):
        # test if domain with np.int32 can be serialized
        p = Continuous("x", domain=[0, np.int32(1)])
        json.dumps(p.to_config())

    def test_with_units(self):
        p = Continuous("x", domain=[1, 10], unit="°C")
        conf = p.to_config()
        assert conf["unit"] == "°C"

    def test_round(self):
        p = Continuous(name="x", domain=[1, 10])
        # scalar
        assert p.round(0) == 1
        assert p.round(3.3) == 3.3
        assert p.round(100) == 10
        # np.array
        assert np.allclose(p.round(np.zeros(10)), 1)
        # pd.Series
        assert np.allclose(p.round(pd.Series([0, 0, 0])), 1)

        # parameter without bounds
        p = Continuous(name="x")
        assert p.round(-10) == -10

    def test_contains(self):
        p = Continuous(name="x", domain=[1, 10])
        # scalar
        assert p.contains(2)
        # np.array
        assert np.all(p.contains(np.array([2.3, 3, 4, 5, 6])))
        # pd.Series
        assert np.all(p.contains(pd.Series([2.3, 3, 4, 5, 6])))

    def test_sample(self):
        p = Continuous(name="x", domain=[1, 10])
        assert np.all(p.contains(p.sample()))

        p = Continuous(name="x", domain=[-np.inf, np.inf])
        assert np.all(p.contains(p.sample()))

        p = Continuous(name="x", domain=[0, np.inf])
        assert np.all(p.contains(p.sample()))

        p = Continuous(name="x", domain=[-np.inf, 0])
        assert np.all(p.contains(p.sample()))

    def test_repr(self):
        p1 = Continuous(name="x", domain=[1, 10])
        p2 = eval(p1.__repr__())
        assert p1.to_config() == p2.to_config()

    def test_fixed(self):
        p = Continuous("x", domain=[7, 7])
        assert p.bounds == (7, 7)
        assert np.allclose(p.sample(), 7)

    def test_unbounded(self):
        p = Continuous("x", domain=[-np.inf, np.inf])
        assert p.contains(p.sample()).all()

        p = Continuous("x", domain=[None, None])
        assert p.contains(p.sample()).all()

        p = Continuous("x")
        assert p.contains(p.sample()).all()

    def test_unit_range(self):
        p = Continuous(name="x", domain=[-10, 20])
        x = np.linspace(-10, 20)
        xt = p.to_unit_range(x)
        assert np.allclose(xt, np.linspace(0, 1))

        p = Continuous(name="x", domain=[1, 1])  # fixed -> don't do anything
        x = np.linspace(-10, 20)
        xt = p.to_unit_range(x)
        assert np.allclose(xt, x)


class TestDiscrete:
    def test_init(self):
        p = Discrete(domain=[1, 2, 3, 5], name="x")
        assert p.name == "x"
        assert p.bounds == (1, 5)
        assert p.domain == [1, 2, 3, 5]

        p = Discrete(domain=[-1.1, 5.1], name="x")
        assert p.bounds == (-1.1, 5.1)
        assert p.domain == [-1.1, 5.1]

    def test_config(self):
        p = Discrete(domain=[1, 2, 3, 5], name="x")
        conf = p.to_config()
        assert conf["type"] == "discrete"
        assert conf["name"] == "x"
        assert conf["domain"] == [1, 2, 3, 5]
        json.dumps(conf)

    def test_config_int32(self):
        # test if domain with np.int32 can be serialized
        p = Discrete("x", domain=np.array([0, 1, 2], dtype=np.int32))
        json.dumps(p.to_config())

    def test_checks(self):
        with pytest.raises(ValueError):
            Discrete("x", domain="very discrete")  # domain must be a list
        with pytest.raises(ValueError):
            Discrete("x", domain=[])  # domain needs to have at least 2 values
        with pytest.raises(ValueError):
            Discrete("x", domain=[1, 1])  # domain cannot contain duplicates
        with pytest.raises(ValueError):
            Discrete("x", domain=["a", "b"])  # domain may only contain numerics

    def test_with_units(self):
        p = Discrete("x", domain=[1, 2, 7, 10], unit="°C")
        conf = p.to_config()
        assert conf["unit"] == "°C"

    def test_is_integer(self):
        assert Discrete(name="x", domain=[0, 1, 2]).is_integer()
        assert Discrete(name="x", domain=[-1, 0, 1]).is_integer()
        assert not Discrete(name="x", domain=[1, 2, 10]).is_integer()
        assert not Discrete(name="x", domain=[1, 2.1, 3]).is_integer()
        assert not Discrete(name="x", domain=[1.1, 2.1]).is_integer()

    def test_round(self):
        p = Discrete(name="x", domain=[1, 2, 3, 10])
        # scalars
        assert p.round(0) == 1
        assert p.round(2) == 2
        assert p.round(100) == 10
        # np.array
        assert np.allclose(p.round(np.zeros(10)), 1)
        # pd.Series
        assert np.allclose(p.round(pd.Series([0, 0, 0])), 1)

        # should also work for non-integer domains
        p = Discrete(domain=[1.1, 2.5, 3.141], name="x")
        assert p.round(0) == 1.1
        assert p.round(4) == 3.141

    def test_contains(self):
        p = Discrete(name="x", domain=[1, 2, 3, 10])
        # scalar
        assert p.contains(2)
        assert not p.contains(4)
        # array
        assert np.all(p.contains([1, 1, 1, 2, 2, 2]))

    def test_sample(self):
        p = Discrete(name="x", domain=[1, 2, 3, 5])
        assert np.all(p.contains(p.sample()))

    def test_repr(self):
        p1 = Discrete(name="x", domain=[1, 2, 3, 5])
        p2 = eval(p1.__repr__())
        assert p1.to_config() == p2.to_config()

    def test_fixed(self):
        p = Discrete("x", domain=[7])
        assert p.bounds == (7, 7)
        assert np.allclose(p.sample(), 7)

    def test_unit_range(self):
        p = Discrete(name="x", domain=[-10, 5, 20])
        x = np.linspace(-10, 20)
        xt = p.to_unit_range(x)
        assert np.allclose(xt, np.linspace(0, 1))

        p = Discrete(name="x", domain=[1])  # fixed -> don't do anything
        x = np.linspace(-10, 20)
        xt = p.to_unit_range(x)
        assert np.allclose(xt, x)


class TestCategorical:
    def test_init(self):
        p = Categorical(domain=["A", "B", 3], name="x")
        assert p.name == "x"
        assert p.domain == ["A", "B", 3]

    def test_checks(self):
        with pytest.raises(ValueError):
            Categorical("x", domain="very categorical")  # domain must be a list
        with pytest.raises(ValueError):
            Categorical("x", domain=[])  # domain needs to have at least 2 values
        with pytest.raises(ValueError):
            Categorical("x", domain=["a"])  # domain needs to have 2+ values
        with pytest.raises(ValueError):
            Categorical("x", domain=["a", "a"])  # domain cannot contain duplicates

    def test_config(self):
        p = Categorical(domain=["A", "B", 3], name="x")
        conf = p.to_config()
        assert conf["type"] == "categorical"
        assert conf["name"] == "x"
        assert conf["domain"] == ["A", "B", 3]
        json.dumps(conf)

    def test_contains(self):
        p = Categorical(name="x", domain=["A", "B", 3])
        assert p.contains("A")
        assert p.contains(["A", "A", 3]).all()
        assert not p.contains("x")
        assert not p.contains(["A", "x"]).all()

    def test_sample(self):
        p = Categorical(name="x", domain=["A", "B", 3])
        x = p.sample()
        assert p.contains(x)

    def test_repr(self):
        p1 = Categorical(name="x", domain=["A", "B", 3])
        p2 = eval(p1.__repr__())
        assert p1.to_config() == p2.to_config()

    def test_round(self):
        p = Categorical(name="x", domain=["A", "B", 3])
        assert p.round("A") == "A"
        x = ["A", "A", "B", 3]
        assert np.all(p.round(x) == x)

    def test_onehot_encoding(self):
        p = Categorical(name="x", domain=["B", "A", "C"])
        points = pd.Series(["A", "A", "C", "B"])
        transformed = p.to_onehot_encoding(points)
        assert np.allclose(transformed["x§A"], [1, 1, 0, 0])
        assert np.allclose(transformed["x§B"], [0, 0, 0, 1])
        assert np.allclose(transformed["x§C"], [0, 0, 1, 0])

        untransformed = p.from_onehot_encoding(transformed)
        assert np.all(points == untransformed)

    def test_dummy_encoding(self):
        p = Categorical(name="x", domain=["B", "A", "C"])
        points = pd.Series(["A", "A", "C", "B"])
        transformed = p.to_dummy_encoding(points)
        # the first level "B" is dropped
        assert np.allclose(transformed["x§A"], [1, 1, 0, 0])
        assert np.allclose(transformed["x§C"], [0, 0, 1, 0])

        untransformed = p.from_dummy_encoding(transformed)
        assert np.all(points == untransformed)

    def test_label_encoding(self):
        p = Categorical(name="x", domain=["B", "A", "C"])
        points = pd.Series(["A", "A", "C", "B"])
        transformed = p.to_label_encoding(points)
        assert np.allclose(transformed, [1, 1, 2, 0])

        untransformed = p.from_label_encoding(transformed)
        assert np.all(points == untransformed)


class TestParameters:
    mixed_parameters = Parameters(
        [
            Categorical(name="category", domain=["city bike", "race bike"]),
            Continuous(name="weight", domain=[5, 20]),
            Discrete(name="gears", domain=[3, 8, 16, 21, 24, 27]),
        ]
    )

    def test_init(self):
        # init from a list/tuple of dimensions
        Parameters(
            (
                Continuous(name="foo", domain=[1, 10]),
                Discrete(name="bar", domain=[1, 2, 3, 4]),
                Categorical(name="baz", domain=["A", "B", 3]),
            )
        )

        # init from a list/tuple of configs
        Parameters(
            [
                {"name": "foo", "type": "continuous", "domain": [1, 10]},
                {"name": "bar", "type": "discrete", "domain": [1, 2, 3, 4]},
                {"name": "baz", "type": "categorical", "domain": ["A", "B", 3]},
            ]
        )

        # not a list/tuple
        with pytest.raises(TypeError):
            Parameters(Continuous(name="weight", domain=[5, 20]))

    def test_config(self):
        # test serializing to config
        conf = self.mixed_parameters.to_config()
        params = Parameters(conf)
        assert params.to_config() == conf

    def test_repr(self):
        # repr uniquely identifies the space
        params = eval(self.mixed_parameters.__repr__())
        assert params.to_config() == self.mixed_parameters.to_config()

    def test_iter(self):
        # test iterating over the dimensions of the parameter space
        n = len([d for d in self.mixed_parameters])
        assert len(self.mixed_parameters) == n

    def test_subscriptable(self):
        # test accessing the dimensions by name
        for name in self.mixed_parameters.names:
            assert self.mixed_parameters[name].name == name

    def test_contains(self):
        # test if points are correctly identified as contained inside the parameter space
        points = pd.DataFrame(
            columns=self.mixed_parameters.names,
            data=[["race bike", 9.12, 8], ["city bike", 16.8, 8]],
        )
        assert self.mixed_parameters.contains(points).all()

    def test_sample(self):
        # sample some points from the input space and check if they are inside the bounds
        points = self.mixed_parameters.sample(3)
        assert self.mixed_parameters.contains(points).all()
        assert points.shape == (3, len(self.mixed_parameters))

        params = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(5)])
        points = params.sample(3)
        assert set(points.columns) == set(["x0", "x1", "x2", "x3", "x4"])
        assert points.shape == (3, len(params))
        assert params.contains(points).all()

        points = params.sample(1)
        assert points.shape == (1, len(params))
        assert params.contains(points).all()

    def test_round(self):
        # test rounding values
        points = pd.DataFrame(
            {
                "category": ["city bike", "race bike", "race bike", "race bike"],
                "weight": [4, 7, 100, 25],
                "gears": [1, 5, 18, 21],
            },
            index=[4, 5, 6, 7],
        )
        rounded = self.mixed_parameters.round(points)
        assert self.mixed_parameters.contains(rounded).all()

        # test rounding for unbouned parameters
        parameters = Parameters([Continuous("x1"), Continuous("x2", [0, 10])])
        points = pd.DataFrame(
            {
                "x1": [-1, 2, 999],
                "x2": [-1, 2, 999],
            }
        )
        rounded = parameters.round(points)
        assert np.allclose(rounded["x1"], [-1, 2, 999])
        assert np.allclose(rounded["x2"], [0, 2, 10])

    def test_bounds(self):
        # test parameter bounds
        params = Parameters([Continuous(f"x{i}", domain=[0, 1]) for i in range(5)])
        assert (params.bounds.loc["min"] == 0).all()
        assert (params.bounds.loc["max"] == 1).all()

        params = Parameters([Discrete(f"x{i}", domain=[0, 1]) for i in range(5)])
        assert (params.bounds.loc["min"] == 0).all()
        assert (params.bounds.loc["max"] == 1).all()

        # no bounds when space includes categorical parameters
        with pytest.raises(TypeError):
            params = Parameters(self.mixed_parameters)
            params.bounds

    def test_transform(self):
        # transform to the unit-range
        params = Parameters([Continuous(f"x{i}", domain=[-10, 10]) for i in range(5)])
        X = params.sample(20)
        Xt = params.transform(X, continuous="normalize")
        assert np.all(Xt >= 0) and np.all(Xt <= 1)

        # transform with mixed parameters
        params = Parameters(
            [
                Continuous("x1", domain=[-10, 10]),
                Discrete("x2", domain=[0, 1, 5]),
                Categorical("x3", domain=["A", "B", "C"]),
            ]
        )
        X = params.sample(20)

        # one-hot
        Xt = params.transform(
            X, continuous="normalize", discrete="normalize", categorical="onehot-encode"
        )
        assert Xt.shape == (20, 5)
        assert np.all(Xt >= 0) and np.all(Xt <= 1)

        # label-encode
        Xt = params.transform(X, categorical="label-encode")
        assert Xt.shape == (20, 3)

        # unkown transforms
        with pytest.raises(ValueError):
            params.transform(X, continuous="foo")
        with pytest.raises(ValueError):
            params.transform(X, discrete="foo")
        with pytest.raises(ValueError):
            params.transform(X, categorical="foo")

    def test_duplicate_names(self):
        # test handling of duplicate parameter names
        with pytest.raises(ValueError):
            Parameters([Continuous("x"), Continuous("x")])

    def test_add(self):
        params1 = Parameters([Continuous(f"x{i}") for i in range(3)])
        params2 = Parameters([Continuous(f"y{i}") for i in range(3)])
        both = params1 + params2
        assert isinstance(both, Parameters)
        assert len(both) == 6



"""
Simple 1D problems for visually assessing probabilistic surrogate models.
Note: these problems should be output-noisified, e.g.
```
import opti

problem = opti.problems.noisify_problem_with_gaussian(
    opti.problems.Line1D(),
    sigma=0.1
)
```
"""
import numpy as np
import pandas as pd

from opti.parameter import Continuous, Discrete
from opti.problem import Problem


class Line1D(Problem):
    def __init__(self):
        def f(x):
            return 0.1 * x + 1

        x = np.array([0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5])
        data = pd.DataFrame({"x": x, "y": f(x)})

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=data,
        )


class Parabola1D(Problem):
    def __init__(self):
        def f(x):
            return 0.025 * (x - 5) ** 2 + 1

        x = np.array([0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5])
        data = pd.DataFrame({"x": x, "y": f(x)})

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=data,
        )


class Sinus1D(Problem):
    def __init__(self):
        def f(x):
            return np.sin(x * 2 * np.pi / 10) / 2 + 2

        x = np.array([0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5])
        data = pd.DataFrame({"x": x, "y": f(x)})

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=data,
        )


class Sigmoid1D(Problem):
    def __init__(self):
        def f(x):
            return 1 / (1 + np.exp(-(x - 5) * 2)) + 1

        x = np.array([0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5])
        data = pd.DataFrame({"x": x, "y": f(x)})

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=data,
        )


class Step1D(Problem):
    def __init__(self):
        def f(x):
            return (x > 1.1).astype(float)

        x = np.array([0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5])
        data = pd.DataFrame({"x": x, "y": f(x)})

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Discrete("y", [0, 1])],
            f=f,
            data=data,
        )



from typing import Callable, Optional, Sequence

import numpy as np

from opti import Problem


def create_multi_fidelity(
    problem: Problem,
    fidelities: Sequence[Callable[[np.ndarray], np.ndarray]],
    predicates: Sequence[Callable[[int, np.ndarray], bool]],
    name: Optional[str] = None,
):
    """
    Based on a given problem, this creates a new problem with different fidelities.
    Internally, a counter is increased on every call.
    Args:
        problem: base problem
        fidelities: callables that evaluate the problem with different fidelities
        predicates: callables that decide which fidelity is active at the current call
                    depending on the counter and the inputs
        name: name of the new multi-fidelity instance

    Returns: new problem with different fidelities

    Example:

        # We apply two Gaussians as two fidelities to the zdt1-problem.

        zdt1 = opti.problems.ZDT1(n_inputs=n_inputs)
        first_fidelity = noisify_problem_with_gaussian(zdt1, mu=0, sigma=0.01)
        second_fidelity = noisify_problem_with_gaussian(zdt1, mu=0.1, sigma=0.5)

        def pred_second(counter, X):
            return counter % 3 == 0

        def pred_first(counter, X):
            return not pred_second(counter, _)

        mf_zdt = create_multi_fidelity(
            zdt1, [first_fidelity.f, second_fidelity.f], [pred_first, pred_second]
        )

        [...]

        # when calling subsequently we obtain

        mf_zdt.f(X)  # first fidelity
        mf_zdt.f(X)  # first fidelity
        mf_zdt.f(X)  # second fidelity
        mf_zdt.f(X)  # first fidelity
        mf_zdt.f(X)  # first fidelity
        mf_zdt.f(X)  # second fidelity
        mf_zdt.f(X)  # first fidelity
        ...


    """

    class MultiFidelityF:
        def __init__(self):
            self.counter = 0

        def __call__(self, X):
            self.counter += 1
            for fidelity, pred in zip(fidelities, predicates):
                if pred(self.counter - 1, X):
                    return fidelity(X)
            raise ValueError("no predicate returned true")

    return Problem(
        inputs=problem.inputs,
        outputs=problem.outputs,
        objectives=problem.objectives,
        constraints=problem.constraints,
        output_constraints=problem.output_constraints,
        f=MultiFidelityF(),
        data=None,
        name=problem.name if name is None else name,
    )



import numpy as np
import pandas as pd

import opti
from opti.constraint import NonlinearInequality
from opti.parameter import Continuous
from opti.problem import Problem


class Qapi1(Problem):
    """Constrained problem from the Qriteria API tests.
    Note that while the function is convex, the constraints are not.

    minimize
        f1(x) = (x1 - 2)^2 + (x2 - 1)^2
        f2(x) = x1^2 + (x2 - 3)^2
    for
        x1 in [0, inf)
        x2 in (-inf, inf)
    s.t.
        0 <= x1
        c1(x) = - x1^2 + x2 <= 0
        c2(x) = - x1 - x2 + 2 <= 0

    The ideal point is [0, 0] and the nadir is [8, 8].
    """

    def __init__(self):
        def f(x):
            x = np.atleast_2d(x).T
            f0 = (x[0] - 2) ** 2 + (x[1] - 1) ** 2
            f1 = x[0] ** 2 + (x[1] - 3) ** 2
            return np.stack([f0, f1], axis=1)

        super().__init__(
            name="Constrained bi-objective problem",
            inputs=[Continuous("x1", [0, 10]), Continuous("x2", [-10, 10])],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(2)],
            constraints=[
                NonlinearInequality("x2 - x1**2"),
                NonlinearInequality("2 - x1 - x2"),
            ],
            f=f,
        )


class Hyperellipsoid(Problem):
    """Hyperellipsoid in n dimensions

    minimize
        f_m(x) = x_m    m = 1, ... n
    for
        x in R^n
    s.t.
        sum((x / a)^2) - 1 <= 0

    The ideal point is -a and the is nadir 0^n.

    Args:
        n (int, optional): Dimension of the hyperellipsoid. Defaults to 5.
        a (list-like, optional): Half length of principal axes. a = None or a = [1, ...] results in a hypersphere.
    """

    def __init__(self, n=5, a=None):
        if a is None:
            a = np.ones(n)
            constr = " + ".join([f"x{i}**2" for i in range(n)]) + " - 1"
        else:
            a = np.array(a).squeeze()
            if len(a) != n:
                raise ValueError("Dimension of half axes doesn't match input dimension")
            constr = " + ".join([f"(x{i}/{a[i]})**2" for i in range(n)]) + " - 1"
        self.a = a

        super().__init__(
            name="Hyperellipsoid",
            inputs=[Continuous(f"x{i}", [-a[i], a[i]]) for i in range(n)],
            outputs=[Continuous(f"y{i}", [-a[i], a[i]]) for i in range(n)],
            constraints=[NonlinearInequality(constr)],
            f=lambda x: np.atleast_2d(x),
        )

    def get_optima(self, n=10) -> pd.DataFrame:
        X = opti.sampling.sphere.sample(self.n_inputs, n, positive=True)
        X = np.concatenate([-np.eye(self.n_inputs), -X], axis=0)[:n]
        Y = self.a * X
        return pd.DataFrame(
            data=np.column_stack([X, Y]),
            columns=self.inputs.names + self.outputs.names,
        )


class Daechert1(Problem):
    """Problem with a non-convex Pareto front.

    The ideal point is [-1.37, -1.61, -4] and the nadir is [0, 0, -1.44].
    """

    def __init__(self):
        def f(x):
            x1, x2, x3 = np.atleast_2d(x).T
            return -np.stack([x1, x2, x3 ** 2], axis=1)

        super().__init__(
            name="Dächert-1",
            inputs=[
                Continuous("x1", domain=[0, np.pi]),
                Continuous("x2", domain=[0, 10]),
                Continuous("x3", domain=[1.2, 10]),
            ],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(3)],
            constraints=[NonlinearInequality("- cos(x1) - exp(-x2) + x3")],
            f=f,
        )


class Daechert2(Problem):
    """Unconstrained problem with a Pareto front resembling a comet.

    minimize
        f1(x) = (1 + x3) (x1^3 x2^2 - 10 x1 - 4 x2)
        f2(x) = (1 + x3) (x1^3 x2^2 - 10 x1 + 4 x2)
        f3(x) = 3 (1 + x3) x1^2
    s.t.
         1 <= x1 <= 3.5
        -2 <= x2 <= 2
         0 <= x3 <= 1

    The ideal point is [-70.19, -70.19, 3] and the nadir is [4, 4, 73.5].
    """

    def __init__(self):
        def f(x):
            x1, x2, x3 = np.atleast_2d(x).T
            return np.stack(
                [
                    (1 + x3) * (x1 ** 3 * x2 ** 2 - 10 * x1 - 4 * x2),
                    (1 + x3) * (x1 ** 3 * x2 ** 2 - 10 * x2 + 4 * x2),
                    3 * (1 + x3) * x1 ** 2,
                ],
                axis=1,
            )

        super().__init__(
            name="Dächert-2",
            inputs=[
                Continuous("x1", domain=[1, 3.5]),
                Continuous("x2", domain=[-2, 2]),
                Continuous("x3", domain=[0, 1]),
            ],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(3)],
            f=f,
        )


class Daechert3(Problem):
    """Modification of DTLZ7, with a Pareto consisting of 4 disconnected parts.

    The ideal point is [0, 0, 2.61] and the nadir is [0.86, 0.86, 6].
    """

    def __init__(self):
        def f(x):
            x = np.atleast_2d(x)
            f0 = x[:, 0]
            f1 = x[:, 1]
            f2 = 6 - np.sum(x * (1 + np.sin(3 * np.pi * x)), axis=1)
            return np.stack([f0, f1, f2], axis=1)

        super().__init__(
            name="Dächert-3",
            inputs=[Continuous(f"x{i}", domain=[0, 1]) for i in range(2)],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(3)],
            f=f,
        )



import json

import numpy as np
import pandas as pd

import opti
from opti.objective import CloseToTarget, Maximize, Minimize, Objectives, make_objective


def make_dataframe():
    return pd.DataFrame(
        {
            "meetings": [10, 21, 10, 1, 2],
            "coffee": [2, 3, 5, 7, 1],
            "seriousness": [5, 8, 2, 4, 10],
        },
        index=["monday", "tuesday", "wednesday", "thursday", "friday"],
    )


def test_minimize():
    obj = Minimize("meetings")
    eval(obj.__repr__())
    json.dumps(obj.to_config())
    make_objective(**obj.to_config())
    assert obj.name == "meetings"

    df = make_dataframe()
    y = df["meetings"].values
    assert np.allclose(obj(y), y)
    assert np.allclose(obj.eval(df), y)


def test_maximize():
    obj = Maximize("coffee")
    eval(obj.__repr__())
    json.dumps(obj.to_config())
    make_objective(**obj.to_config())
    assert obj.name == "coffee"

    df = make_dataframe()
    y = df["coffee"].values
    assert np.allclose(obj(y), -y)
    assert np.allclose(obj.eval(df), -y)


def test_closetotarget():
    obj = CloseToTarget("seriousness", target=5, exponent=2)
    eval(obj.__repr__())
    json.dumps(obj.to_config())
    make_objective(**obj.to_config())
    assert obj.name == "seriousness"

    df = make_dataframe()
    y = df["seriousness"].values
    assert np.allclose(obj(y), (y - 5) ** 2)
    assert np.allclose(obj.eval(df), (y - 5) ** 2)


def test_objectives():
    objectives = Objectives(
        [
            Minimize("meetings", target=2),
            Maximize("coffee", target=4),
            CloseToTarget("seriousness", target=5, exponent=2, tolerance=1.1),
        ]
    )

    objectives.__repr__()
    json.dumps(objectives.to_config())

    assert objectives.names == [
        "meetings",
        "coffee",
        "seriousness",
    ]

    df = make_dataframe()
    Z = objectives.eval(df)
    assert len(Z) == len(df)


def test_ideal_nadir():
    # problem with maximize objectives
    problem = opti.problems.Detergent()
    bounds = problem.objectives.bounds(problem.outputs)
    assert np.allclose(bounds.loc["min"], -3)
    assert np.allclose(bounds.loc["max"], 0)

    # problem with close-to-target objective
    problem = opti.problems.Cake()
    bounds = problem.objectives.bounds(problem.outputs)
    assert np.allclose(bounds.loc["min"], [300, -5, 0])
    assert np.allclose(bounds.loc["max"], [600, 0, 1.4])



import numpy as np

import opti
from opti.metric import pareto_front


def test_pareto_efficient():
    # case 1:
    x = np.linspace(0, 10, 1000)
    A = np.zeros((1000, 2))
    A[:, 0] = (x - 2) ** 2
    A[:, 1] = 3 * x

    efficient = opti.metric.is_pareto_efficient(A)

    # check pareto-efficient points
    for a in A[efficient]:
        # only the point itself is allowed to be strictly equal or better on all objectives
        assert np.all(A <= a, axis=1).sum() == 1
        # all other pareto points must be equal or better in at least one objective
        assert np.any(A[efficient] <= a, axis=1).sum() == efficient.sum()

    # check dominated points
    for a in A[~efficient]:
        # point cannot be better than any pareto optimal point in any objective
        assert np.any(A[efficient] <= a, axis=1).any()

    # case 2:
    A = np.array(
        [
            [0.2, 0.6],
            [0.1, 0.7],
            [0.0, 1.3],
            [0.5, 0.3],
            [0.0, 3.0],
            [0.0, 3.1],
            [5.0, 5.0],
        ]
    )
    efficient = opti.metric.is_pareto_efficient(A)
    assert len(A[efficient]) == 4
    assert np.alltrue(pareto_front(A) == A[efficient])


def test_crowding_distance():
    # concrete example
    # - for the two outer points the crowding distance should be inf (missing neighbor)
    # - for the inner two points 0.7 is the average cuboide side length and 0.9 is the
    #   scaling from the spread in each dimension
    A = np.array([[0.1, 1], [0.3, 0.8], [0.8, 0.3], [1, 0.1]])
    np.testing.assert_array_almost_equal(
        opti.metric.crowding_distance(A),
        np.array([np.inf, 0.7 / 0.9, 0.7 / 0.9, np.inf]),
    )


def test_generational_distance():
    # Test case: the reference front forms a unit quarter circle around (1, 1)
    φ = np.linspace(0, np.pi / 2, 100)
    R = 1 - np.array([np.sin(φ), np.cos(φ)]).T
    A = np.array([[0.1, 1], [0.3, 0.8], [0.8, 0.3], [1, 0.1]])

    # single point on the front
    assert opti.metric.generational_distance(np.array([[1, 0]]), R) == 0
    # single point at distance = 0.1 to the front
    assert opti.metric.generational_distance(np.array([[1, 0.1]]), R) == 0.1
    # multiple points
    assert opti.metric.generational_distance(A, R) == 0.18603015676967713


def test_inverted_generational_distance():
    # Test case: the reference front forms a unit quarter circle around (1, 1)
    φ = np.linspace(0, np.pi / 2, 100)
    R = 1 - np.array([np.sin(φ), np.cos(φ)]).T
    A = np.array([[0.1, 1], [0.3, 0.8], [0.8, 0.3], [1, 0.1]])

    # single point at distance = 1 to the front
    assert opti.metric.inverted_generational_distance(np.array([[1, 1]]), R) == 1
    # multiple points
    assert opti.metric.inverted_generational_distance(A, R) == 0.292415441318418



import pandas as pd

from opti import Problem
from opti.constraint import (
    Constraints,
    LinearEquality,
    LinearInequality,
    NChooseK,
    NonlinearEquality,
    NonlinearInequality,
)
from opti.objective import CloseToTarget, Maximize, Minimize, Objectives
from opti.parameter import Categorical, Continuous, Discrete, Parameters


def test_overview():
    # inputs, outputs
    inputs = Parameters(
        [
            Continuous("x1", domain=[0, 1]),
            Continuous("x2", domain=[0, 1]),
            Continuous("x3", domain=[0, 1]),
            Discrete("x4", domain=[1, 2, 5, 7.5]),
            Categorical("x5", domain=["A", "B", "C"]),
        ]
    )

    outputs = Parameters(
        [
            Continuous("y1", domain=[0, None]),
            Continuous("y2", domain=[None, None]),
            Continuous("y3", domain=[0, 100]),
        ]
    )

    X = inputs.sample(5)
    inputs.contains(X)
    inputs.names

    # constraints
    constr1 = LinearEquality(["x1", "x2", "x3"], lhs=[1, 1, 1], rhs=1)
    constr2 = LinearInequality(["x1", "x3"], lhs=[1, 1], rhs=0.8)
    constr3 = NonlinearEquality("x1**2 + x2**2 - 1")
    constr4 = NonlinearInequality("1/x1 + 1/x2 - 2")
    constr5 = NChooseK(["x1", "x2", "x3"], max_active=2)
    constraints = Constraints([constr1, constr2, constr3, constr4, constr5])
    constr2.satisfied(X).values
    constr2.eval(X).values

    # objectives
    objectives = Objectives(
        [Minimize("y1"), Maximize("y2"), CloseToTarget("y3", target=7)]
    )
    Y = pd.DataFrame({"y1": [1, 2, 3], "y2": [7, 4, 5], "y3": [5, 6.9, 12]})
    objectives.eval(Y)

    # problem
    Problem(
        inputs=inputs, outputs=outputs, constraints=constraints, objectives=objectives
    )



from opti import read_modde


def test_read_modde():
    problem = read_modde("examples/bread.mip")
    assert problem.data is not None
    assert problem.inputs.names == [
        "kneading",
        "yeast_type",
        # "yeast_age", uncontrolled parameters not yet supported
        "flour",
        "water",
        "salt",
        "yeast",
        "kneading_time",
        "resting_time",
        "baking_time",
        "baking_temperature",
    ]
    assert problem.outputs.names == [
        "crust",
        "elasticity",
        "softness_24h",
        "taste",
        "appearance",
    ]



import json

import numpy as np
import pandas as pd
import pytest

from opti.constraint import (
    Constraints,
    LinearEquality,
    LinearInequality,
    NChooseK,
    NonlinearEquality,
    NonlinearInequality,
    make_constraint,
)


def test_linear_equality():
    names = [f"x{i}" for i in range(5)]
    constraint = LinearEquality(names=names, lhs=np.ones(5), rhs=5)

    df = pd.DataFrame([[1, 1, 1, 1, 1], [1, 2, 3, 4, 5]], columns=names)
    assert np.allclose(constraint.eval(df), [0, 10])
    assert np.allclose(constraint.satisfied(df), [True, False])

    eval(constraint.__repr__())
    json.dumps(constraint.to_config())
    constraint2 = make_constraint(**constraint.to_config())
    assert np.allclose(constraint2.eval(df), constraint.eval(df))

    constraint = LinearEquality(names=names)
    assert np.allclose(constraint.lhs, 1)
    assert constraint.rhs == 0

    with pytest.raises(ValueError):
        LinearEquality(names=names, lhs=[1, 1, 1])


def test_linear_inequality():
    names = [f"x{i}" for i in range(5)]
    constraint = LinearInequality(names=names, lhs=np.ones(5), rhs=5)

    df = pd.DataFrame([[1, 0.1, 1, 1, 1], [1, 2, 3, 4, 5]], columns=names)
    assert np.allclose(constraint.eval(df), [-0.9, 10])
    assert np.allclose(constraint.satisfied(df), [True, False])

    eval(constraint.__repr__())
    json.dumps(constraint.to_config())
    constraint2 = make_constraint(**constraint.to_config())
    assert np.allclose(constraint2.eval(df), constraint.eval(df))

    constraint = LinearInequality(names=names)
    assert np.allclose(constraint.lhs, 1)
    assert constraint.rhs == 0

    with pytest.raises(ValueError):
        LinearInequality(names=names, lhs=[1, 1, 1])


def test_nonlinear_equality():
    constraint = NonlinearEquality("x1**2 + x2**2 - 1")
    df = pd.DataFrame(
        {
            "x1": [0.6, 0.5, 1],
            "x2": [0.8, 0.5, 1],
            "x3": [1, 1, 3],
        }
    )

    assert np.allclose(constraint.eval(df), [0, -0.5, 1])
    assert np.allclose(constraint.satisfied(df), [True, False, False])

    eval(constraint.__repr__())
    json.dumps(constraint.to_config())
    constraint2 = make_constraint(**constraint.to_config())
    assert np.allclose(constraint2.eval(df), constraint.eval(df))


def test_nonlinear_inequality():
    constraint = NonlinearInequality("x1**2 + x2**2 - 1")
    df = pd.DataFrame(
        {
            "x1": [0.6, 0.5, 1],
            "x2": [0.8, 0.5, 1],
            "x3": [1, 1, 3],
        }
    )

    assert np.allclose(constraint.eval(df), [0, -0.5, 1])
    assert np.allclose(constraint.satisfied(df), [True, True, False])

    eval(constraint.__repr__())
    json.dumps(constraint.to_config())
    constraint2 = make_constraint(**constraint.to_config())
    assert np.allclose(constraint2.eval(df), constraint.eval(df))


def test_nchoosek():
    names = ["x1", "x2", "x3", "x4"]
    constraint = NChooseK(names=names, max_active=3)

    df = pd.DataFrame([[1, 0, 1, 1], [1, 2, 4, 4]], columns=names)

    assert np.allclose(constraint.satisfied(df), [True, False])

    eval(constraint.__repr__())
    json.dumps(constraint.to_config())
    make_constraint(**constraint.to_config())


def test_constraints():
    constraints = Constraints(
        [
            LinearEquality(["x1", "x2", "x3"], lhs=[1, 1, 1], rhs=1),
            LinearInequality(["x1", "x2"], lhs=[1, 1], rhs=0.5),
            NonlinearEquality("x1 + x2 + x3 - 1"),
            NonlinearInequality("x1**2 + x3**2 - 1"),
            NChooseK(["x1", "x2"], max_active=2),
        ]
    )

    df = pd.DataFrame({"x1": [0.1, 0.3], "x2": [0.2, 0.3], "x3": [0.7, 0.4]})
    assert np.allclose(constraints.satisfied(df), [True, False])

    # serialize
    config = constraints.to_config()
    constraints = Constraints(config)
    assert np.allclose(constraints.satisfied(df), [True, False])

    # skip empty constraints
    config = [{"type": "linear-equality", "names": [], "lhs": []}]
    constraints = Constraints(config)
    assert len(constraints) == 0


def test_make_constraint():
    with pytest.raises(ValueError):
        make_constraint(type="box-bound")  # unknown constraint type



import pprint
from typing import Dict, List, Sequence, Union

import numpy as np
import pandas as pd


class Constraint:
    """Base class to define constraints on the input space, g(x) = 0 or g(x) <= 0."""

    def eval(self, data: pd.DataFrame) -> pd.Series:
        """Numerically evaluate the constraint.

        Args:
            data: Data to evaluate the constraint on.

        Returns:
            Constraint evaluation g(x) with equalities interpreted as g(x) = 0 and inequalities as g(x) <=0.
        """
        raise NotImplementedError

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        """Check if a constraint is satisfied.

        Args:
            data: Data to evaluate the constraint on.

        Returns:
            Constraint evaluation g(x) = 0 or g(x) <= 0 depending on the constraint type.
        """
        raise NotImplementedError

    def to_config(self) -> Dict:
        raise NotImplementedError


class Constraints:
    """List of optimization constraints"""

    def __init__(self, constraints: Sequence):
        self.constraints = []
        for c in constraints:
            if not isinstance(c, Constraint):
                if "names" in c and len(c["names"]) == 0:
                    continue  # skip empty constraints
                c = make_constraint(**c)
            self.constraints.append(c)

    def __repr__(self):
        return "Constraints(\n" + pprint.pformat(self.constraints) + "\n)"

    def __iter__(self):
        return iter(self.constraints)

    def __len__(self):
        return len(self.constraints)

    def __getitem__(self, i):
        return self.constraints[i]

    def eval(self, data: pd.DataFrame) -> pd.DataFrame:
        """Numerically evaluate all constraints.

        Args:
            data: Data to evaluate the constraints on.

        Returns:
            Constraint evaluation g(x) for each of the constraints.
        """
        return pd.concat([c.eval(data) for c in self.constraints], axis=1)

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        """Check if all constraints are satisfied.

        Args:
            data: Data to evaluate the constraints on.

        Returns:
            Series of booleans indicating if all constraints are satisfied.
        """
        return pd.concat([c.satisfied(data) for c in self.constraints], axis=1).all(
            axis=1
        )

    def to_config(self) -> List[Dict]:
        return [obj.to_config() for obj in self.constraints]


class LinearEquality(Constraint):
    def __init__(
        self, names: List[str], lhs: Union[float, np.ndarray] = 1, rhs: float = 0
    ):
        """Linear / affine inequality of the form 'lhs * x == rhs'.

        Args:
            names: Parameter names that the constraint works on.
            lhs: Left-hand side / coefficients of the constraint.
            rhs: Right-hand side of the constraint.

        Examples:
            A mixture constraint where A, B and C need to add up to 100 can be defined as
            ```
            LinearEquality(["A", "B", "C"], rhs=100)
            ```
            If the coefficients of A, B and C are not 1 they are passed explicitly.
            ```
            LinearEquality(["A", "B", "C"], lhs=[10, 2, 5], rhs=100)
            ```
        """
        self.names = names
        if np.isscalar(lhs):
            self.lhs = lhs * np.ones(len(names))
        else:
            self.lhs = np.asarray(lhs)
        if self.lhs.shape != (len(names),):
            raise ValueError("Number of parameters and coefficients/lhs don't match.")
        self.rhs = rhs
        self.is_equality = True

    def eval(self, data: pd.DataFrame) -> pd.Series:
        return data[self.names] @ self.lhs - self.rhs

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        return pd.Series(np.isclose(self.eval(data), 0), index=data.index)

    def __repr__(self):
        return (
            f"LinearEquality(names={self.names}, lhs={list(self.lhs)}, rhs={self.rhs})"
        )

    def to_config(self) -> Dict:
        return dict(
            type="linear-equality",
            names=self.names,
            lhs=self.lhs.tolist(),
            rhs=self.rhs,
        )


class LinearInequality(Constraint):
    def __init__(
        self, names: List[str], lhs: Union[float, np.ndarray] = 1, rhs: float = 0
    ):
        """Linear / affine inequality of the form 'lhs * x <= rhs'.

        Args:
            names: Parameter names that the constraint works on.
            lhs: Left-hand side / coefficients of the constraint.
            rhs: Right-hand side of the constraint.

        Examples:
            A mixture constraint where the values of A, B and C may not exceed 100 can be defined as
            ```
            LinearInequality(["A", "B", "C"], rhs=100)
            ```
            If the coefficients are not 1, they need to be passed explicitly.
            ```
            LinearInequality(["A", "B", "C"], lhs=[10, 2, 5], rhs=100)
            ```
            Inequalities are alway of the form g(x) <= 0. To define a the constraint g(x) >=0 0, both `lhs` and `rhs` need to be multiplied by -1.
            ```
            LinearInequality(["A", "B", "C"], lhs=-1, rhs=-100)
            LinearInequality(["A", "B", "C"], lhs=[-10, -2, -5], rhs=-100)
            ```
        """
        self.names = names
        if np.isscalar(lhs):
            self.lhs = lhs * np.ones(len(names))
        else:
            self.lhs = np.asarray(lhs)
        if self.lhs.shape != (len(names),):
            raise ValueError("Number of parameters and coefficients/lhs don't match.")
        self.rhs = rhs
        self.is_equality = False

    def eval(self, data: pd.DataFrame) -> pd.Series:
        return data[self.names] @ self.lhs - self.rhs

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        return self.eval(data) <= 0

    def __repr__(self):
        return f"LinearInequality(names={self.names}, lhs={list(self.lhs)}, rhs={self.rhs})"

    def to_config(self) -> Dict:
        return dict(
            type="linear-inequality",
            names=self.names,
            lhs=self.lhs.tolist(),
            rhs=self.rhs,
        )


class NonlinearEquality(Constraint):
    def __init__(self, expression: str):
        """Equality of the form 'expression == 0'.

        Args:
            expression: Mathematical expression that can be evaluated by `pandas.eval`.

        Examples:
            You can pass any expression that can be evaluated by `pd.eval`.
            To define x1**2 + x2**2 = 1, use
            ```
            NonlinearEquality("x1**2 + x2**2 - 1")
            ```
            Standard mathematical operators are supported.
            ```
            NonlinearEquality("sin(A) / (exp(B) - 1)")
            ```
            Parameter names with special characters or spaces need to be enclosed in backticks.
            ```
            NonlinearEquality("1 - `weight A` / `weight B`")
            ```
        """
        self.expression = expression
        self.is_equality = True

    def eval(self, data: pd.DataFrame) -> pd.Series:
        return data.eval(self.expression)

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        return pd.Series(np.isclose(self.eval(data), 0), index=data.index)

    def __repr__(self):
        return f"NonlinearEquality('{self.expression}')"

    def to_config(self) -> Dict:
        return dict(type="nonlinear-equality", expression=self.expression)


class NonlinearInequality(Constraint):
    def __init__(self, expression: str):
        """Inequality of the form 'expression <= 0'.

        Args:
            expression: Mathematical expression that can be evaluated by `pandas.eval`.

        Examples:
            You can pass any expression that can be evaluated by `pd.eval`.
            To define x1**2 + x2**2 < 1, use
            ```
            NonlinearInequality("x1**2 + x2**2 - 1")
            ```
            Standard mathematical operators are supported.
            ```
            NonlinearInequality("sin(A) / (exp(B) - 1)")
            ```
            Parameter names with special characters or spaces need to be enclosed in backticks.
            ```
            NonlinearInequality("1 - `weight A` / `weight B`")
            ```
        """
        self.expression = expression
        self.is_equality = False

    def eval(self, data: pd.DataFrame) -> pd.Series:
        return data.eval(self.expression)

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        return self.eval(data) <= 0

    def __repr__(self):
        return f"NonlinearInequality('{self.expression}')"

    def to_config(self) -> Dict:
        return dict(type="nonlinear-inequality", expression=self.expression)


class NChooseK(Constraint):
    def __init__(self, names: List[str], max_active: int):
        """Only k out of n values are allowed to take nonzero values.

        Args:
            names: Parameter names that the constraint works on.
            max_active: Maximium number of non-zero parameter values.

        Examples:
            A choice of 2 or less from A, B, C, D or E can be defined as
            ```
            NChooseK(["A", "B", "C", "D", "E"], max_active=2)
            ```
        """
        self.names = names
        self.max_active = max_active
        self.is_equality = False

    def eval(self, data: pd.DataFrame) -> pd.Series:
        x = data[self.names].values
        num_zeros = x.shape[1] - self.max_active
        violation = np.apply_along_axis(
            func1d=lambda r: sum(sorted(r)[:num_zeros]), axis=1, arr=x
        )
        return pd.Series(violation, index=data.index)

    def satisfied(self, data: pd.DataFrame) -> pd.Series:
        return pd.Series(self.eval(data) <= 0, index=data.index)

    def __repr__(self):
        return f"NChooseK(names={self.names}, max_active={self.max_active})"

    def to_config(self) -> Dict:
        return dict(type="n-choose-k", names=self.names, max_active=self.max_active)


def make_constraint(type, **kwargs):
    t = type.lower()
    if t == "linear-equality":
        return LinearEquality(**kwargs)
    if t == "linear-inequality":
        return LinearInequality(**kwargs)
    if t == "nonlinear-equality":
        return NonlinearEquality(**kwargs)
    if t == "nonlinear-inequality":
        return NonlinearInequality(**kwargs)
    if t == "n-choose-k":
        return NChooseK(**kwargs)
    raise ValueError(f"Unknown constraint type: {t}")



import json

import numpy as np
import pandas as pd
import pytest

import opti
from opti.model import LinearModel
from opti.objective import Minimize
from opti.parameter import Continuous
from opti.problem import Problem


def test_parameters():
    # test if inputs / outputs can be specified as list of Dimensions
    problem = opti.Problem(
        inputs=[opti.Continuous(f"x{i}", domain=[0, 1]) for i in range(3)],
        outputs=[opti.Continuous(f"y{i}", domain=[0, 1]) for i in range(3)],
    )
    assert isinstance(problem.inputs, opti.Parameters)
    assert isinstance(problem.outputs, opti.Parameters)

    # test if inputs / outputs can be specified as list of Dimensions
    problem = opti.Problem(
        inputs=[
            {"name": "x1", "type": "continuous", "domain": [1, 10]},
            {"name": "x2", "type": "categorical", "domain": ["A", "B", 3]},
        ],
        outputs=[
            {"name": "y1", "type": "continuous", "domain": [1, 10]},
            {"name": "y2", "type": "discrete", "domain": [1, 2, 3]},
        ],
    )
    assert isinstance(problem.inputs, opti.Parameters)
    assert isinstance(problem.outputs, opti.Parameters)


def test_problem():
    # test loading from json
    problem = opti.read_json("examples/bread.json")
    assert len(problem.inputs) == 11
    assert len(problem.outputs) == 4
    assert problem.constraints is not None
    assert problem.name is not None

    data = pd.DataFrame(
        columns=problem.inputs.names + problem.outputs.names,
        data=np.zeros((3, len(problem.inputs) + len(problem.outputs))),
    )
    data["yeast_type"] = "A"
    problem.set_data(data)

    # test repr/str
    problem.__str__()
    problem.__repr__()

    # objectives are optional and should be set to minimize if not given
    config = json.load(open("examples/bread.json"))
    config.pop("objectives")
    problem = opti.Problem.from_config(config)
    assert len(problem.objectives) == len(problem.outputs)
    for obj in problem.objectives:
        assert isinstance(obj, opti.objective.Minimize)

    # constraints are optional
    config = json.load(open("examples/bread.json"))
    config.pop("constraints")
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None


def test_inconsistent_parameters():
    # duplicate parameter name in inputs and outputs
    with pytest.raises(ValueError):
        opti.Problem(inputs=[opti.Continuous("x")], outputs=[opti.Continuous("x")])


def test_inconsistent_objectives():
    # parameter name in objective not among inputs & outputs
    with pytest.raises(ValueError):
        opti.Problem(
            inputs=[opti.Continuous("x")],
            outputs=[opti.Continuous("y")],
            objectives=[opti.objective.Minimize("z")],
        )


def test_set_data():
    problem = opti.Problem(
        inputs=[opti.Discrete("x", [0, 1, 2])], outputs=[opti.Continuous("y")]
    )
    problem.set_data(data=pd.DataFrame({"x": [0, 1], "y": [0, 1]}))

    # missing input
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"y": [0, 1, 2]}))

    # missing output
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1, 2]}))

    # missing value in inputs
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1, np.nan], "y": [0, 1, 2]}))

    # no value in outputs
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1], "y": [np.nan, np.nan]}))

    # non-numeric value for continuous parameter
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": ["oops", 1, 2], "y": [0, 1, 2]}))
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": [0, 1, 2], "y": ["oops", 1, 2]}))

    # test with categorical parameter
    problem = opti.Problem(
        inputs=[opti.Categorical("x", ["A", "B"])], outputs=[opti.Continuous("y")]
    )

    problem.set_data(data=pd.DataFrame({"x": ["A", "A"], "y": [0, 1]}))

    # unknown category
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": ["A", "B", "C"], "y": [0, 1, 2]}))


def test_add_data():
    problem = opti.Problem(
        inputs=[opti.Continuous("x")], outputs=[opti.Continuous("y")]
    )

    # adding data when no data exists
    problem.add_data(data=pd.DataFrame({"x": [0, 1], "y": [0, 1]}))
    assert len(problem.data) == 2

    # adding new data with extra column
    problem.add_data(data=pd.DataFrame({"x": [2], "y": [2], "z": [2]}))
    assert len(problem.data) == 3
    assert "z" in problem.data.columns

    # adding data with missing input value
    with pytest.raises(ValueError):
        problem.add_data(data=pd.DataFrame({"x": [np.nan], "y": [3]}))

    # adding data with missing input value
    with pytest.raises(ValueError):
        problem.add_data(data=pd.DataFrame({"x": [3]}))


def test_get_XY():
    problem = opti.Problem(
        inputs=[
            opti.Continuous("x1", [0, 1]),
            opti.Discrete("x2", [0, 1, 5]),
            opti.Categorical("x3", ["A", "B", "C"]),
        ],
        outputs=[opti.Continuous("y1"), opti.Continuous("y2")],
        data=pd.DataFrame(
            {
                "x1": [0.2, 0.5, 0.6],
                "x2": [0, 1, 5],
                "x3": ["A", "C", "B"],
                "y1": [1, 0.8, 0.4],
                "y2": [0, 0.2, np.nan],
            }
        ),
    )

    X = problem.get_X()
    assert X.shape == (3, 3)
    Y = problem.get_Y()
    assert Y.shape == (3, 2)

    # nans are dropped
    X, Y = problem.get_XY()
    assert X.shape == (2, 3)
    assert Y.shape == (2, 2)

    # using transforms
    X, Y = problem.get_XY(
        continuous="normalize", discrete="normalize", categorical="onehot-encode"
    )
    assert X.shape == (2, 5)
    assert Y.shape == (2, 2)

    # consider subset of outputs
    X, Y = problem.get_XY(outputs=["y1"])
    assert X.shape == (3, 3)
    assert Y.shape == (3, 1)

    # pass data
    data = pd.DataFrame(
        {
            "x1": [0, 0, 0],
            "x2": [0, 0, 0],
            "x3": ["A", "B", "C"],
            "y1": [1, np.nan, 0.4],
            "y2": [0, 0.2, np.nan],
        }
    )
    X, Y = problem.get_XY(data=data)
    assert X.shape == (1, 3)
    assert Y.shape == (1, 2)


def test_X_bounds():
    # Obtaining input data bounds should work for continuous and discrete parameters.
    # For parameters with 0 variation, require xhi = xlo + 1.
    problem = opti.Problem(
        inputs=[
            opti.Continuous("x1", [-1, 1]),
            opti.Continuous("x2", [-1, 1]),
            opti.Discrete("x3", [0, 1, 2]),
        ],
        outputs=[opti.Continuous("y")],
        data=pd.DataFrame({"x1": [-0.2, 0.5], "x2": [0, 0], "x3": [2, 0], "y": [0, 1]}),
    )
    xlo, xhi = problem.get_X_bounds()
    assert np.allclose(xlo, [-0.2, 0, 0])
    assert np.allclose(xhi, [0.5, 1, 2])


def test_empty_constraints():
    problem = opti.Problem(
        inputs=[opti.Continuous("x")],
        outputs=[opti.Continuous("y")],
        constraints=[],
    )
    assert problem.constraints is None

    config = {
        "inputs": [{"type": "continuous", "name": "x", "domain": [0, 1]}],
        "outputs": [{"type": "continuous", "name": "y", "domain": [0, 1]}],
        "constraints": [],
    }
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None

    config = {
        "inputs": [{"type": "continuous", "name": "x"}],
        "outputs": [{"type": "continuous", "name": "y"}],
        "constraints": [
            {"names": [], "lhs": [], "rhs": 1, "type": "linear-inequality"}
        ],
    }
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None


def test_eval():
    problem = opti.problems.ZDT1(3)
    X = problem.inputs.sample(2)
    Y = problem.eval(X)
    assert len(Y) == 2
    assert (X.index == Y.index).all()


def test_config():
    # test if the configuration dict is strictly json-compliant
    problem = opti.problems.Ackley()
    problem.create_initial_data(2)
    problem.data.loc[0, "y"] = np.nan

    conf = problem.to_config()
    # domain [-np.inf, np.inf] should be converted to [None, None]
    assert conf["outputs"][0]["domain"][0] is None
    assert conf["outputs"][0]["domain"][1] is None
    # the the datapoint y = np.nan should be converted to None
    assert conf["data"]["data"][0][-1] is None


def test_json(tmpdir):
    tmpfile = tmpdir.join("foo.json")

    # test if problems are serializable to/from json
    problem = opti.Problem.from_json("examples/bread.json")
    problem.to_json(tmpfile)

    # test json compliance
    problem = opti.problems.Ackley()  # output domain is [-Inf, Inf]
    problem.to_json(tmpfile)

    conf = json.load(open(tmpfile))
    assert conf["outputs"][0]["domain"][0] is None
    assert conf["outputs"][0]["domain"][1] is None

    problem = opti.Problem.from_json(tmpfile)
    assert problem.outputs.bounds.loc["min", "y"] == -np.inf
    assert problem.outputs.bounds.loc["max", "y"] == np.inf

    # test unicode characters
    problem = opti.Problem(
        inputs=[opti.Continuous("öäüßèé", [0, 1])],
        outputs=[opti.Continuous("/[].#@²³$")],
    )
    problem.to_json(tmpfile)
    problem2 = opti.read_json(tmpfile)
    assert np.all(problem2.inputs.names == problem.inputs.names)
    assert np.all(problem2.outputs.names == problem.outputs.names)


def test_optima():
    problem = opti.read_json("examples/simple.json")
    assert len(problem.optima) == 2


def test_models():
    D = 5
    M = 3
    problem = Problem(
        inputs=[Continuous(f"x{i}", [0, 1]) for i in range(D)],
        outputs=[Continuous(f"y{i}") for i in range(M)],
        objectives=[Minimize(f"y{i}") for i in range(M - 1)],
        models=[LinearModel([f"y{i}"], coefficients=np.ones(D)) for i in range(M)],
    )

    X = problem.sample_inputs(10)
    Y = problem.models.eval(X)
    Z = problem.objectives.eval(Y)
    assert list(Y.columns) == problem.outputs.names
    assert list(Z.columns) == problem.objectives.names



import json

import numpy as np
import pandas as pd
from sklearn.cross_decomposition import PLSRegression
from sklearn.datasets import make_regression

from opti.model import LinearModel, Model, Models, make_model


def test_linear_model():
    model = LinearModel(["y"], [1, 2, 3], 10)
    assert model.names == ["y"]
    eval(model.__repr__())

    # de-/serialize
    json.dumps(model.to_config())
    make_model(**model.to_config())

    # evaluate,
    X = pd.DataFrame(np.random.rand(10, 3), columns=["x1", "x2", "x3"])
    y = model.eval(X)
    assert y.columns == ["y"]
    assert len(y) == 10


def test_models():
    model1 = LinearModel(["cost"], [1, 2, 3])
    model2 = LinearModel(["pcf"], [5, 2.1, 0.9])
    models = Models([model1, model2])

    assert models[0] == model1
    assert models[1] == model2
    assert len(models) == 2
    assert models.names == ["cost", "pcf"]

    json.dumps(models.to_config())
    models = Models(models.to_config())

    # evaluate
    X = pd.DataFrame(np.random.rand(10, 3), columns=["x1", "x2", "x3"])
    Y = models.eval(X)
    assert list(Y.columns) == ["cost", "pcf"]
    assert len(Y) == 10


def test_custom_model():
    D = 5

    # custom sklearn model
    X_train, Y_train = make_regression(
        n_features=D, n_informative=4, n_targets=3, n_samples=20
    )
    pls = PLSRegression(n_components=4)
    pls.fit(X_train, Y_train)
    model1 = Model(names=["y1", "y2", "y3"])
    model1.eval = lambda x: pd.DataFrame(
        pls.predict(x.values), columns=model1.names, index=x.index
    )

    # linear model
    model2 = LinearModel(["cost"], [1, 2, 3, 4, 5])

    models = Models([model1, model2])
    assert models.names == ["y1", "y2", "y3", "cost"]

    # evaluate
    X = pd.DataFrame(np.random.rand(10, D), columns=[f"x{i}" for i in range(D)])
    Y = models.eval(X)
    assert list(Y.columns) == ["y1", "y2", "y3", "cost"]
    assert len(Y) == 10

    # to config, the custom model is omitted
    assert len(models.to_config()) == 1



import warnings
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from scipy.stats.qmc import Sobol

from opti.constraint import Constraints, LinearEquality, NChooseK, NonlinearEquality
from opti.parameter import Continuous, Parameters
from opti.sampling.polytope import polytope_sampling


def split_nchoosek(
    constraints: Optional[Constraints],
) -> Tuple[Constraints, Constraints]:
    """Split constraints in n-choose-k constraint and all other constraints."""
    if constraints is None:
        return Constraints([]), Constraints([])
    nchoosek_constraints = Constraints(
        [c for c in constraints if isinstance(c, NChooseK)]
    )
    other_constraints = Constraints(
        [c for c in constraints if not isinstance(c, NChooseK)]
    )
    return nchoosek_constraints, other_constraints


def apply_nchoosek(samples: pd.DataFrame, constraint: NChooseK):
    """Apply an n-choose-k constraint in-place"""
    n_zeros = len(constraint.names) - constraint.max_active
    for i in samples.index:
        s = np.random.choice(constraint.names, size=n_zeros, replace=False)
        samples.loc[i, s] = 0


def constrained_sampling(
    n_samples: int, parameters: Parameters, constraints: Constraints
) -> pd.DataFrame:
    """Uniform sampling from a constrained space."""
    nchoosek_constraints, other_constraints = split_nchoosek(constraints)

    try:
        samples = rejection_sampling(n_samples, parameters, other_constraints)
    except Exception:
        samples = polytope_sampling(n_samples, parameters, other_constraints)

    if len(nchoosek_constraints) == 0:
        return samples

    for c in nchoosek_constraints:
        apply_nchoosek(samples, c)
    # check if other constraints are still satisfied
    if not constraints.satisfied(samples).all():
        raise Exception(
            "Applying the n-choose-k constraint(s) violated another constraint."
        )
    return samples


def sobol_sampling(n_samples: int, parameters: Parameters) -> pd.DataFrame:
    """Super-uniform sampling from an unconstrained space using a Sobol sequence."""
    d = len(parameters)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X = Sobol(d).random(n_samples)
    res = []
    for i, p in enumerate(parameters):
        if isinstance(p, Continuous):
            x = p.from_unit_range(X[:, i])
        else:
            bins = np.linspace(0, 1, len(p.domain) + 1)
            idx = np.digitize(X[:, i], bins) - 1
            x = np.array(p.domain)[idx]
        res.append(pd.Series(x, name=p.name))
    return pd.concat(res, axis=1)


def rejection_sampling(
    n_samples: int,
    parameters: Parameters,
    constraints: Constraints,
    max_iters: int = 1000,
) -> pd.DataFrame:
    """Uniformly distributed samples from a constrained space via rejection sampling."""
    if constraints is None:
        return parameters.sample(n_samples)

    # check for equality constraints in combination with continuous parameters
    for c in constraints:
        if isinstance(c, (LinearEquality, NonlinearEquality)):
            for p in parameters:
                if isinstance(p, Continuous):
                    raise Exception(
                        "Rejection sampling doesn't work for equality constraints over continuous variables."
                    )

    n_iters = 0
    n_found = 0
    points_found = []
    while n_found < n_samples:
        n_iters += 1
        if n_iters > max_iters:
            raise Exception("Maximum iterations exceeded in rejection sampling")
        points = parameters.sample(10000)
        valid = constraints.satisfied(points)
        n_found += np.sum(valid)
        points_found.append(points[valid])

    return pd.concat(points_found, ignore_index=True).iloc[:n_samples]



import numpy as np
import pandas as pd

from opti.constraint import LinearInequality, NChooseK
from opti.parameter import Categorical, Continuous
from opti.problem import Problem


class Ackley(Problem):
    def __init__(self, n_inputs=2):
        super().__init__(
            name="Ackley function",
            inputs=[Continuous(f"x{i}", [-32.768, +32.768]) for i in range(n_inputs)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x):
        a = 20
        b = 1 / 5
        c = 2 * np.pi
        n = self.n_inputs
        x = np.atleast_2d(x)
        part1 = -a * np.exp(-b * np.sqrt((1 / n) * np.sum(x ** 2, axis=-1)))
        part2 = -np.exp((1 / n) * np.sum(np.cos(c * x), axis=-1))
        return part1 + part2 + a + np.exp(1)

    def get_optima(self):
        x = np.zeros((1, self.n_inputs))
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Himmelblau(Problem):
    def __init__(self):
        super().__init__(
            name="Himmelblau function",
            inputs=[Continuous(f"x{i}", [-6, 6]) for i in range(2)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x):
        x0, x1 = np.atleast_2d(x).T
        return (x0 ** 2 + x1 - 11) ** 2 + (x0 + x1 ** 2 - 7) ** 2

    def get_optima(self):
        x = np.array(
            [
                [3.0, 2.0],
                [-2.805118, 3.131312],
                [-3.779310, -3.283186],
                [3.584428, -1.848126],
            ]
        )
        y = np.zeros(4)
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Rosenbrock(Problem):
    def __init__(self, n_inputs=2):
        super().__init__(
            name="Rosenbrock function",
            inputs=[Continuous(f"x{i}", [-2.048, 2.048]) for i in range(n_inputs)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x):
        x = np.atleast_2d(x).T
        return np.sum(100 * (x[1:] - x[:-1] ** 2) ** 2 + (1 - x[:-1]) ** 2, axis=0)

    def get_optima(self):
        x = np.ones((1, self.n_inputs))
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Schwefel(Problem):
    def __init__(self, n_inputs=2):
        super().__init__(
            name="Schwefel function",
            inputs=[Continuous(f"x{i}", [-500, 500]) for i in range(n_inputs)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x):
        x = np.atleast_2d(x)
        return 418.9829 * self.n_inputs - np.sum(x * np.sin(np.abs(x) ** 0.5), axis=1)

    def get_optima(self):
        x = np.full((1, self.n_inputs), 420.9687)
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Sphere(Problem):
    def __init__(self, n_inputs=10):
        super().__init__(
            name="Sphere function",
            inputs=[Continuous(f"x{i}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous("y", [0, 2])],
        )

    def f(self, x):
        x = np.atleast_2d(x)
        return np.sum((x - 0.5) ** 2, axis=1)

    def get_optima(self):
        x = np.full((1, self.n_inputs), 0.5)
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Rastrigin(Problem):
    def __init__(self, n_inputs=2):
        super().__init__(
            name="Rastrigin function",
            inputs=[Continuous(f"x{i}", [-5, 5]) for i in range(n_inputs)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x):
        x = np.atleast_2d(x)
        a = 10
        return a * self.n_inputs + np.sum(x ** 2 - a * np.cos(2 * np.pi * x), axis=1)

    def get_optima(self):
        x = np.zeros((1, self.n_inputs))
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Zakharov(Problem):
    def __init__(self, n_inputs=2):
        super().__init__(
            name="Zakharov function",
            inputs=[Continuous(f"x{i}", [-10, 10]) for i in range(n_inputs)],
            outputs=[Continuous("y", [-np.inf, np.inf])],
        )

    def f(self, x: np.ndarray):
        x = np.atleast_2d(x)
        a = 0.5 * np.sum(np.arange(1, self.n_inputs + 1) * x, axis=1)
        return np.sum(x ** 2, axis=1) + a ** 2 + a ** 4

    def get_optima(self):
        x = np.zeros((1, self.n_inputs))
        y = 0
        return pd.DataFrame(np.c_[x, y], columns=self.inputs.names + self.outputs.names)


class Zakharov_NChooseKConstraint(Problem):
    """Variant of the Zakharov problem with an n-choose-k constraint"""

    def __init__(self, n_inputs=5, n_max_active=3):
        base = Zakharov(n_inputs)
        self.base = base
        super().__init__(
            name="Zakharov with n-choose-k constraint",
            inputs=base.inputs,
            outputs=base.outputs,
            constraints=[NChooseK(names=base.inputs.names, max_active=n_max_active)],
            f=base.f,
        )

    def get_optima(self):
        return self.base.get_optima()


class Zakharov_Constrained(Problem):
    """Variant of the Zakharov problem with one linear constraint"""

    def __init__(self, n_inputs=5):
        base = Zakharov(n_inputs)
        self.base = base
        super().__init__(
            name="Zakharov with one linear constraint",
            inputs=base.inputs,
            outputs=base.outputs,
            constraints=[LinearInequality(base.inputs.names, lhs=1, rhs=10)],
            f=base.f,
        )

    def get_optima(self):
        return self.base.get_optima()


class Zakharov_Categorical(Problem):
    """Variant of the Zakharov problem with one categorical input"""

    def __init__(self, n_inputs=3):
        base = Zakharov(n_inputs)
        super().__init__(
            name="Zakharov function with one categorical input",
            inputs=[Continuous(f"x{i}", [-10, 10]) for i in range(n_inputs - 1)]
            + [Categorical("expon_switch", ["one", "two"])],
            outputs=base.outputs,
        )

    def f(self, x: np.ndarray):
        x_conti = np.atleast_2d(x[:, :-1])  # Just the continuous inputs
        a = 0.5 * np.sum(np.arange(1, self.n_inputs) * x_conti, axis=1)
        powers = np.repeat(np.expand_dims([2.0, 2.0, 4.0], 0), repeats=len(x), axis=0)
        modify_powers = x[:, -1] == "two"
        powers[modify_powers, :] += powers[modify_powers, :]
        res = (
            np.sum(x_conti ** np.expand_dims(powers[:, 0], 1), axis=1)
            + a ** np.expand_dims(powers[:, 1], 0)
            + a ** np.expand_dims(powers[:, 2], 0)
        )
        res_float_array = np.array(res, dtype=np.float64).ravel()
        return res_float_array

    def get_optima(self):
        x = list(np.zeros(self.n_inputs - 1)) + ["one"]
        y = [0]
        return pd.DataFrame([x + y], columns=self.inputs.names + self.outputs.names)



import pprint
from typing import Dict, List, Union

import numpy as np
import pandas as pd

from opti.parameter import Parameters


class Objective:
    def __init__(self, name: str, parameter: Union[str, List[str]]):
        """Base class for optimzation objectives.

        Args:
            name: name of the objective
            parameters: parameter(s) that the objective is operating on
        """
        self.name = name
        self.parameter = parameter

    def __call__(self, y: np.ndarray) -> np.ndarray:
        """Evaluate the objective values for given output values."""
        raise NotImplementedError  # implemented in the derived classes

    def eval(self, df: pd.DataFrame) -> pd.Series:
        """Evaluate the objective values for a given DataFrame."""
        return pd.Series(
            self(df[self.parameter].values), index=df.index, name=self.name
        )

    def to_config(self) -> Dict:
        """Return a json-serializable dictionary of the objective."""
        raise NotImplementedError  # implemented in the derived classes


class Minimize(Objective):
    def __init__(self, parameter: str, target: float = 0):
        """Minimization objective

        s(y) = y - target

        Args:
            parameter: parameter to minimize
            target: value below which no further improvement is required
        """
        super().__init__(name=parameter, parameter=parameter)
        self.target = target

    def __call__(self, y: np.ndarray) -> np.ndarray:
        return y - self.target

    def untransform(self, y: np.ndarray) -> np.ndarray:
        """Undo the transformation from output to objective value"""
        return y + self.target

    def __repr__(self):
        return f"Minimize('{self.parameter}', target={self.target})"

    def to_config(self) -> Dict:
        config = dict(name=self.parameter, type="minimize")
        if self.target != 0:
            config["target"] = self.target
        return config


class Maximize(Objective):
    def __init__(self, parameter: str, target: float = 0):
        """Maximization objective

        s(y) = target - y

        Args:
            name: name of the objective (= parameter to optimize)
            target: value above which no further improvement is required
        """
        super().__init__(name=parameter, parameter=parameter)
        self.target = target

    def __call__(self, y: np.ndarray) -> np.ndarray:
        return -y + self.target

    def untransform(self, y: np.ndarray) -> np.ndarray:
        """Undo the transformation from output to objective value"""
        return -y - self.target

    def __repr__(self):
        return f"Maximize('{self.parameter}', target={self.target})"

    def to_config(self) -> Dict:
        config = dict(name=self.parameter, type="maximize")
        if self.target != 0:
            config["target"] = str(self.target)
        return config


class CloseToTarget(Objective):
    def __init__(
        self,
        parameter: str,
        target: float = 0,
        exponent: float = 1,
        tolerance: float = 0,
    ):
        """Objective for getting as close as possible to a given value.

        s(y) = |y - target| ** exponent - tolerance ** exponent

        Args:
            parameter: parameter to optimize
            target: target value
            exponent: exponent of the difference
            tolerance: distance to target below which no further improvement is required
        """
        super().__init__(name=parameter, parameter=parameter)
        self.target = target
        self.exponent = exponent
        self.tolerance = tolerance

    def __call__(self, y: np.ndarray) -> np.ndarray:
        return (
            np.abs(y - self.target) ** self.exponent - self.tolerance ** self.exponent
        )

    def __repr__(self):
        return f"CloseToTarget('{self.parameter}', target={self.target})"

    def to_config(self) -> Dict:
        config = dict(name=self.parameter, type="close-to-target", target=self.target)
        if self.exponent != 1:
            config["exponent"] = self.exponent
        if self.tolerance != 0:
            config["tolerance"] = self.tolerance
        return config


class Objectives:
    """Container for optimization objectives.

    Objectives can be either used to quantify the optimility or as a constraint on the
    viability of output values (chance / feasibility constraint)
    """

    def __init__(self, objectives: Union[List[Objective], List[Dict]]):
        _objectives = []
        for m in objectives:
            if isinstance(m, Objective):
                _objectives.append(m)
            else:
                _objectives.append(make_objective(**m))
        self.objectives = _objectives

    def __call__(self, y: np.ndarray) -> np.ndarray:
        y = np.atleast_2d(y)
        return np.stack([obj(yi) for yi, obj in zip(y.T, self.objectives)], axis=-1)

    def eval(self, y: pd.DataFrame) -> pd.DataFrame:
        return pd.concat([obj.eval(y) for obj in self.objectives], axis=1)

    def __repr__(self):
        return "Objectives(\n" + pprint.pformat(self.objectives) + "\n)"

    def __iter__(self):
        return iter(self.objectives)

    def __len__(self):
        return len(self.objectives)

    def __getitem__(self, i: int) -> Objective:
        return self.objectives[i]

    @property
    def names(self):
        return [obj.name for obj in self]

    def bounds(self, outputs: Parameters) -> pd.DataFrame:
        """Compute the bounds in objective space based on the output space bounds.

        The bounds can be interpreted as the ideal and nadir points.

        Examples for continuous parameters:
            min y for y in [0, 10] -> ideal = 0, nadir = 10
            max y for y in [0, 10] -> ideal = -10, nadir = 0
            min (y - 7)**2 for y in [0, 10] -> ideal = 0, nadir = 7**2

        Args:
            outputs: Output parameters.
        """
        Z = self.eval(outputs.bounds)
        bounds = pd.DataFrame(columns=self.names, dtype=float)
        bounds.loc["min"] = Z.min(axis=0)
        bounds.loc["max"] = Z.max(axis=0)

        for name, obj in zip(self.names, self):
            if isinstance(obj, CloseToTarget):
                bounds.loc["min", name] = 0

        return bounds

    def to_config(self) -> List[Dict]:
        return [obj.to_config() for obj in self.objectives]


def make_objective(type: str, name: Union[str, List[str]], **kwargs) -> Objective:
    """Make an objective from a configuration.

    ```
    obj = make_objective(**config)
    ```

    Args:
        type: objective type
        name: parameter(s) that the objective is operating on
    """
    objective = {
        "minimize": Minimize,
        "maximize": Maximize,
        "close-to-target": CloseToTarget,
    }[type.lower()]
    return objective(parameter=name, **kwargs)



"""Specifications and tools for multi-objective optimization problems."""
# flake8: noqa

__version__ = "0.9.4"

from opti import (
    constraint,
    metric,
    model,
    objective,
    parameter,
    problem,
    problems,
    sampling,
    tools,
)
from opti.constraint import (
    Constraints,
    LinearEquality,
    LinearInequality,
    NChooseK,
    NonlinearEquality,
    NonlinearInequality,
    make_constraint,
)
from opti.model import LinearModel, Model, Models, make_model
from opti.objective import CloseToTarget, Maximize, Minimize, Objectives, make_objective
from opti.parameter import Categorical, Continuous, Discrete, Parameters, make_parameter
from opti.problem import Problem, read_json
from opti.tools.modde import read_modde



import numpy as np


def sample(dimension: int, n_samples: int = 1, positive: bool = False) -> np.ndarray:
    """Sample uniformly from the unit hypersphere.

    Args:
        dimension (int): Number of dimensions.
        n_samples (int): Number of samples to draw.
        positive (bool): Sample from the non-negative unit-sphere.

    Returns:
        array, shape=(n_samples, dimesnion): Random samples from the unit simplex.
    """
    x = np.random.normal(0, 1, size=(n_samples, dimension))
    x = x / np.sum(x ** 2, axis=1, keepdims=True) ** 0.5
    if positive:
        x *= -2 * (x < 0) + 1
    return x



import pprint
from typing import Dict, List, Union

import pandas as pd


class Model:
    def __init__(self, names: List[str]):
        """Base class for models of outputs as function of inputs.

        Args:
            names: names of the modeled outputs
        """
        for name in names:
            if not isinstance(name, str):
                ValueError("Model: names must be a list of strings")
        self.names = list(names)

    def eval(self, df: pd.DataFrame) -> pd.DataFrame:
        """Evaluate the objective values for a given DataFrame."""
        raise NotImplementedError

    def to_config(self) -> None:
        """Return a json-serializable dictionary of the objective."""
        pass  # non-serializable models should be ommited without raising an error


class LinearModel(Model):
    def __init__(self, names: List[str], coefficients, offset: float = 0):
        super().__init__(names)
        if len(names) > 1:
            raise ValueError("LinearModel can only describe a single output")
        self.coefficients = coefficients
        self.offset = offset

    def eval(self, x: pd.DataFrame) -> pd.DataFrame:
        y = x @ self.coefficients + self.offset
        return pd.DataFrame(y, columns=self.names)

    def __repr__(self):
        return f"LinearModel({self.names}, coefficients={self.coefficients}, offset={self.offset})"

    def to_config(self) -> Dict:
        return dict(
            type="linear-model",
            names=self.names,
            coefficients=self.coefficients,
            offset=self.offset,
        )


class Models:
    """Container for models."""

    def __init__(self, models: Union[List[Model], List[Dict]]):
        _models = []
        for m in models:
            if isinstance(m, Model):
                _models.append(m)
            else:
                _models.append(make_model(**m))
        self.models = _models

    def eval(self, y: pd.DataFrame) -> pd.DataFrame:
        return pd.concat([model.eval(y) for model in self.models], axis=1)

    def __repr__(self):
        return "Models(\n" + pprint.pformat(self.models) + "\n)"

    def __iter__(self):
        return iter(self.models)

    def __len__(self):
        return len(self.models)

    def __getitem__(self, i: int) -> Model:
        return self.models[i]

    @property
    def names(self):
        names = []
        for model in self.models:
            names += model.names
        return names

    def to_config(self) -> List[Dict]:
        return [
            model.to_config() for model in self.models if model.to_config() is not None
        ]


def make_model(type, **kwargs):
    t = type.lower()
    if t == "linear-model":
        return LinearModel(**kwargs)
    raise ValueError(f"Unknown model type: {t}")



"""
ZDT problem test suite.
Reference: Zitzler, Eckart, Kalyanmoy Deb, and Lothar Thiele. “Comparison of multiobjective evolutionary algorithms: Empirical results.” Evolutionary computation 8.2 (2000): 173-195. doi: 10.1.1.30.5848
Adapted from https://github.com/msu-coinlab/pymoo/blob/master/pymoo/problems/multi/zdt.py
"""
import numpy as np
import pandas as pd

from opti.parameter import Continuous
from opti.problem import Problem


class ZDT1(Problem):
    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT1 function",
            inputs=[Continuous(f"x{i}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i}", [0, np.inf]) for i in range(2)],
        )

    def f(self, x):
        f1 = x[:, 0]
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x[:, 1:], axis=1)
        f2 = g * (1 - (f1 / g) ** 0.5)
        return np.stack([f1, f2], axis=1)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.sqrt(x)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT2(Problem):
    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT2 function",
            inputs=[Continuous(f"x{i}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i}", [0, np.inf]) for i in range(2)],
        )

    def f(self, x):
        f1 = x[:, 0]
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x[:, 1:], axis=1)
        f2 = g * (1 - (f1 / g) ** 2)
        return np.stack([f1, f2], axis=1)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.power(x, 2)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT3(Problem):
    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT3 function",
            inputs=[Continuous(f"x{i}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(2)],
        )

    def f(self, x):
        f1 = x[:, 0]
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x[:, 1:], axis=1)
        f2 = g * (1 - (f1 / g) ** 0.5 - (f1 / g) * np.sin(10 * np.pi * f1))
        return np.stack([f1, f2], axis=1)

    def get_optima(self, points=100):
        regions = [
            [0, 0.0830015349],
            [0.182228780, 0.2577623634],
            [0.4093136748, 0.4538821041],
            [0.6183967944, 0.6525117038],
            [0.8233317983, 0.8518328654],
        ]

        pf = []
        for r in regions:
            x1 = np.linspace(r[0], r[1], int(points / len(regions)))
            x2 = 1 - np.sqrt(x1) - x1 * np.sin(10 * np.pi * x1)
            pf.append(np.stack([x1, x2], axis=1))

        y = np.concatenate(pf, axis=0)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT4(Problem):
    def __init__(self, n_inputs=10):
        super().__init__(
            name="ZDT4 function",
            inputs=[Continuous(f"x{0}", [0, 1])]
            + [Continuous(f"x{i}", [-5, 5]) for i in range(1, n_inputs)],
            outputs=[Continuous(f"y{i}", [0, np.inf]) for i in range(2)],
        )

    def f(self, x):
        f1 = x[:, 0]
        g = 1 + 10 * (self.n_inputs - 1)
        for i in range(1, self.n_inputs):
            g += x[:, i] ** 2 - 10 * np.cos(4.0 * np.pi * x[:, i])
        f2 = g * (1 - np.sqrt(f1 / g))
        return np.stack([f1, f2], axis=1)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.sqrt(x)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT6(Problem):
    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT6 function",
            inputs=[Continuous(f"x{i}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i}", [-np.inf, np.inf]) for i in range(2)],
        )

    def f(self, x):
        n = self.n_inputs
        f1 = 1 - np.exp(-4 * x[:, 0]) * (np.sin(6 * np.pi * x[:, 0])) ** 6
        g = 1 + 9 * (np.sum(x[:, 1:], axis=1) / (n - 1)) ** 0.25
        f2 = g * (1 - (f1 / g) ** 2)
        return np.stack([f1, f2], axis=1)

    def get_optima(self, points=100):
        x = np.linspace(0.2807753191, 1, points)
        y = np.stack([x, 1 - x ** 2], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)
